package sv.data;

import java.awt.*;
import java.io.*;
import java.util.*;

/** 
 * Define normal point data in a given time
 */

public class TDPoint extends TimeData1D {

  // to hold AMR color values
  private Color[] colorArr;

  // to hold error values
  private float[] errArr = null;

  /**
   * constructor
   * @param time time value
   * @param noOfPoints the number of points of this data
   * @param xArr x value array
   * @param yArr y value array
   */
  public TDPoint(TimeValue time, int noOfPoints, float[] xArr, float[] yArr) {
     super(time, noOfPoints, xArr, yArr);
     dataType = TimeData1D.POINT;
     rank = ONED;
  }

  /**
   * constructor
   * @param time time value
   * @param noOfPoints the number of points of this data
   * @param xArr x value array
   * @param yArr y value array
   * @param err error value array
   */
  public TDPoint(TimeValue time, int noOfPoints, float[] xArr, float[] yArr, 
                     float[] err, boolean dummy) {

     super(time, noOfPoints, xArr, yArr);
     dataType = TimeData1D.POINT;
     rank = ONED;
     errArr = err;
     recalcMinMax();
  }

  /**
   * constructor
   * @param time time value
   * @param noOfPoints the number of points of this data
   * @param xArr x value array
   * @param yArr y value array
   * @param cArr color value array
   */
  public TDPoint(TimeValue time, int noOfPoints, float[] xArr, float[] yArr, float[] cArr) {
     super(time, noOfPoints, xArr, yArr, cArr);
     dataType = TimeData1D.POINT;
     rank = ONED;
  }

  /**
   * constructor
   * @param time time value
   * @param noOfPoints the number of points of this data
   * @param xArr x value array
   * @param yArr y value array
   * @param err error value array
   * @param cArr color value array
   */
  public TDPoint(TimeValue time, int noOfPoints, float[] xArr, float[] yArr, 
                                                   float[] err, float[] cArr) {
     super(time, noOfPoints, xArr, yArr, cArr);
     dataType = TimeData1D.POINT;
     rank = ONED;
     errArr = err;
     recalcMinMax();
  }

  /**
   * constructor by using bounding box
   * @param time time value
   * @param noOfPoints the number of points of this data
   * @param xmin x min value
   * @param xmin x max value
   * @param yArr y value array
   */
  public TDPoint(TimeValue time, int noOfPoints, 
                     float xmin, float xmax, float[] yArr) {
     super(time, noOfPoints, xmin, xmax, yArr);
     dataType = TimeData1D.POINT;
     rank = ONED;
  }

  /**
   * constructor by using bounding box
   * @param time time value
   * @param noOfPoints the number of points of this data
   * @param xmin x min value
   * @param xmin x max value
   * @param yArr y value array
   * @param err error value array
   */
  public TDPoint(TimeValue time, int noOfPoints, 
                     float xmin, float xmax, float[] yArr, float[] err) {
     super(time, noOfPoints, xmin, xmax, yArr);
     dataType = TimeData1D.POINT;
     rank = ONED;
     errArr = err;
     recalcMinMax();
  }

  /** 
   * sending data by using DataOutputStream
   */
  public void sendingData(DataOutputStream out) throws IOException {

    // send the magic number
    out.writeInt(101);

    // send time
    out.writeFloat(time.getAbsTime());

    // send number of points in given time
    out.writeInt(noOfPoints);
     
    // send x and y arrays
    if (bb) {
      out.writeByte(1);     // bb is true... send boundbox first
      out.writeFloat(xmin);
      out.writeFloat(xmax);

      for (int i=0; i<noOfPoints; i++) 
        out.writeFloat(yArr[i]);

    } else {
      out.writeByte(0);      // bb is false

      for (int i=0; i<noOfPoints; i++) {
        out.writeFloat(xArr[i]);
        out.writeFloat(yArr[i]);
      }
    }

    // send cArr
    if (cArr == null) {
      out.writeByte(0);
    } else {
      out.writeByte(1);
      for (int i=0; i<noOfPoints; i++) 
         out.writeFloat(cArr[i]);
    }

    // send error array
    if (errArr == null) {
      out.writeByte(0);
    } else {
      out.writeByte(1);
      for (int i=0; i<noOfPoints; i++) 
         out.writeFloat(errArr[i]);
    }

    // send reserved for future use
    out.writeInt(0);
  }

  /**
   * return data size of this time data
   */
  public long getDataSize() {
    long sz = 4;         // initially time value
    
    if (xArr != null) sz = sz + 4 * xArr.length;
    if (yArr != null) sz = sz + 4 * yArr.length;
    if (cArr != null) sz = sz + 4 * cArr.length;
    if (colorArr != null) sz = sz + 4 * colorArr.length;
    if (errArr != null) sz = sz + 4 * errArr.length;
    return sz;
  }

  /**
   * return error array
   */
  public float[] getErrorArr() {
     return errArr;
  }

  private void recalcMinMax() {
     if (errArr == null) return;

     float emax = Float.MIN_VALUE;

     for (int i=0; i< noOfPoints; i++) 
        if (emax < errArr[i]) emax = errArr[i];

     ymin = ymin - emax;
     ymax = ymax + emax;
  }

} // end of class


