package sv.engine;

import java.awt.*;
import java.awt.event.*;
import java.net.*;
import java.io.*;
import java.util.*;
import com.sun.java.swing.*;
import com.sun.java.swing.event.*;
import sv.engine.SV2Engine;
import sv.data.*;
import sv.viewer.GFrame;
import sv.viewer.builtin.*;

/** 
 * cRecvHandler
 * work after getting connection from client or collaboration server
 */

public class cRecvHandler extends Thread {

  private Socket sock;
  private DataOutputStream out;
  private DataInputStream in;

  private SV2Engine engine;

  /**
   * constructor
   * @param s server socket
   * @param eng parent module(SV2Engine)
   */
   public cRecvHandler(Socket s, SV2Engine eng) {
      sock = s;
      engine = eng;
   }

   /**
    * read commands from the client and do work
    */
   public void run() {
     try {
       out = new DataOutputStream(sock.getOutputStream());
       in  = new DataInputStream(new 
                        BufferedInputStream(sock.getInputStream()));

       String hostname = sock.getInetAddress().getHostName();
       // engine.getGUI().message("Connection from: " + hostname, false);
 
       int command = in.readInt();

       /////////////////////////////////////////////////////////////////
       // Now we have to decide what commands and data are received 
       // from the client. Normally, after receiving data, connection will
       // be disconnected. 

       doWork(hostname, command);       // do the server work
 
       out.close();                     // close streams and socket
       in.close();
       sock.close();                   

       //////////////////////////////////////////////////////////////////

     } catch (Exception e) {
       e.printStackTrace();
     }
  }


  // real worker
  private synchronized void doWork(String hostname, int command) {
    switch(command) {
      case SV2Protocol.HEAD_DATA_FEED:  
        doHEAD_DATA_FEED();
        break;

      case SV2Protocol.NEW_MASTER:
        doNEW_MASTER();
        break;

      case SV2Protocol.BROADCAST_COMMAND:
        byte[] datum = null;
        try {
          datum = new byte[in.readInt()];
          in.readFully(datum);
        } catch(Exception e) { }
        engine.vmanager.doDataWork(datum);
        break;

      case SV2Protocol.ACCESS_DENIED:
        engine.getGUI().message("(Slave)Server denied the access.", true);
        break;

      case SV2Protocol.PROTOCOL_ERROR:
        engine.getGUI().message("Protocol error.", true);
        break;

      case SV2Protocol.COOKED_DATA_FEED:
        doCOOKED_DATA_FEED();
        break;

      case SV2Protocol.VOLUME_IMAGE_FEED:
        doVOLUME_IMAGE_FEED();
        break;

      case SV2Protocol.VOLUME_IMAGE_BULK:
        doVOLUME_IMAGE_BULK();
        break;

      case SV2Protocol.CLIENT_FILTER_WORK:
        doCLIENT_FILTER_WORK();
        break;

      case SV2Protocol.CACHE_TRANSFER:
        doCACHE_TRANSFER();
        break;

      default:
        System.err.println("Got undefined command ... " + command);
     }
  }

  // received brief data information
  private void doHEAD_DATA_FEED() {
    DataSetBrief brief = null;

    engine.getGUI().message("Receiving brief data information.", false);
    try {
      String title = in.readUTF();
      int noOfTime= in.readInt();
      int rank = in.readInt();
      long dataSize = in.readLong();
      long id = in.readLong();

      brief = new DataSetBrief(title, noOfTime, rank, dataSize);
      brief.setID(id);

    } catch (Exception e) {
      engine.getGUI().message("Comm error: " + SV2Protocol.HEAD_DATA_FEED, true);
      return;
    }
 
    engine.receivedBriefInfo(brief);   
  }

  // this client is now new master
  private void doNEW_MASTER() {
    JOptionPane.showMessageDialog(engine.getGUI(), 
        "This client is elected as a new master.", 
        "New Master", JOptionPane.INFORMATION_MESSAGE);

    engine.sender.setMaster(true);
  }

  // receive data set
  private void doCOOKED_DATA_FEED() {

    long vid = -1;     // viewer id
    long did = -1;     // data set id

    engine.getGUI().message("Receiving data from the server.", false);
    engine.getGUI().progress(0);

    try {
      // read the viewer id
      vid = in.readLong();
  
      // read the data set id
      did = in.readLong();
  
      // reaceive the number of data set
      int noOfDataSet = in.readInt();
      DataSet[] dataSets = new DataSet[noOfDataSet];

      double percent = 0.0;
      double datastep1 = 100.0/noOfDataSet;

      // find the data set with the given id and add received real data
      DataSetBrief brief = engine.getGUI().iconDisplay.getDataSetBrief(did);
      if (brief == null) {
        engine.getGUI().message("Missing data set error.", true);
        return;
      }

      for (int i=0; i<noOfDataSet; i++) {
        String title = in.readUTF();

        // read the number of time data
        int noOfTimeData = in.readInt(); 
        TimeData[] readed = new TimeData[noOfTimeData];

        double datastep = datastep1/noOfTimeData;

        for (int k=0; k<noOfTimeData; k++) {
          int type = in.readInt();

          switch (type) {
   	    case 101:
              readed[k] = DataReader.parse101(in);
              break;
   	    case 102:
              readed[k] = DataReader.parse102(in);
              break;
   	    case 103:
              readed[k] = DataReader.parse103(in);
              break;
   	    case 104:
              readed[k] = DataReader.parse104(in);
              break;
   	    case 105:
              readed[k] = DataReader.parse105(in);
              break;
            case 201:
              readed[k] = DataReader.parse201(in);
              break;
   	    case 302:
              readed[k] = DataReader.parse302(in);
              break;

  	    default:
              System.out.println("Received not defined data.");
              return;
          } 
          percent += datastep;
          engine.getGUI().progress((int)percent);
        }
        dataSets[i] = new DataSet(readed);
        brief.addDataSet(dataSets[i]);
      }

      brief.updateInfo();
      brief.setViewerID(vid); 

      // check viewer id 
      GFrame gframe = engine.vmanager.isExist(vid);
      if (gframe != null) 
        gframe.updateDataSet();
      else 
        engine.vmanager.register(brief, vid, null);  // register it

      engine.getGUI().progress(0);

    } catch (Exception e) {
e.printStackTrace();
      engine.getGUI().message("Error while receiving data.", true);
      return;
    }
    engine.getGUI().message("Receiving: done.", false);
  }
  
  // receive volume image
  private void doVOLUME_IMAGE_FEED() {
    long vid = -1;     // viewer id
    long did = -1;     // data set id
    DataSet dataSet = null;

    engine.getGUI().message("Receiving volume image from the server.", false);
    engine.getGUI().progress(0);

    try {
      // read the viewer id
      vid = in.readLong();
  
      // read the data set id
      did = in.readLong();

      // read the volume image
      int sz = in.readInt();
      double[] vars = new double[sz];
      for (int i=0; i<sz; i++)
        vars[i] = in.readDouble();
      engine.getGUI().progress(50);
    
      byte[] arr = new byte[3*256*256];
      in.readFully(arr);
      engine.getGUI().progress(100);

      // find the data set with the given id and add received real data
      DataSetBrief brief = engine.getGUI().iconDisplay.getDataSetBrief(did);
      if (brief == null) {
        engine.getGUI().message("Missing data set error.", true);
        return;
      }
      brief.setViewerID(vid); 
     
      // check viewer id 
      GFrame gframe = engine.vmanager.isExist(vid);
      if (gframe != null) 
        ((GFrameVol)gframe).updateImage(engine.getCacheDir(), arr, vars, true);
      else 
        engine.vmanager.register(brief, engine.getCacheDir(), arr, vars, vid); 
      
      engine.getGUI().progress(0);

    } catch (Exception e) {
e.printStackTrace();
      engine.getGUI().message("Error while receiving data.", true);
      return;
    }
    engine.getGUI().message("Receiving: done.", false);
  }


  // receive bulk volume image
  private void doVOLUME_IMAGE_BULK() {
    long vid = -1;     // viewer id
    long did = -1;     // data set id
    DataSet dataSet = null;
    int bulk_sz = 0;

    engine.getGUI().message("Receiving bulk volume image from the server.", false);
    engine.getGUI().progress(0);

    try {
      // read the viewer id
      vid = in.readLong();
  
      // read the data set id
      did = in.readLong();

      // find the data set with the given id and add received real data
      DataSetBrief brief = engine.getGUI().iconDisplay.getDataSetBrief(did);
      if (brief == null) {
        engine.getGUI().message("Missing data set error.", true);
        return;
      }
      // check viewer id 
      GFrame gframe = engine.vmanager.isExist(vid);
      if (gframe == null) {
        System.out.println("Missing frame error.");
        return;
      }
     
      bulk_sz = in.readInt();
      double percent_step = 100.0 / bulk_sz;

      // read the volume image
      for (int m=0; m<bulk_sz; m++) {
        int sz = in.readInt();
        double[] vars = new double[sz];
        for (int i=0; i<sz; i++)
          vars[i] = in.readDouble();
        engine.getGUI().progress((int)(percent_step * (double)m));
    
        byte[] arr = new byte[3*256*256];
        in.readFully(arr);
        ((GFrameVol)gframe).updateImage(engine.getCacheDir(), arr, vars, false);
      }

      ((GFrameVol)gframe).receivedBulkImages();
      engine.getGUI().progress(0);

    } catch (Exception e) {
e.printStackTrace();
      engine.getGUI().message("Error while receiving data.", true);
      return;
    }
    engine.getGUI().message("Receiving: done.", false);
  }


  // client filter work
  private void doCLIENT_FILTER_WORK() {
    long id1, id2;
    String filter_name, argument;
    byte[] filter_code;
    boolean overlay, discard;
    long did = -1, vid = -1;

    try {
      id1 = in.readLong();
      filter_name = in.readUTF();
      int sz = in.readInt();
      filter_code = new byte[sz];
      in.readFully(filter_code);
      argument  = in.readUTF();
      id2 = in.readLong();
      overlay = Boolean.valueOf(in.readUTF()).booleanValue();
      discard = Boolean.valueOf(in.readUTF()).booleanValue();

      int newdata = in.readInt();
      if (newdata == 1) {
        did = in.readLong();
        vid = in.readLong();
      }
      engine.vmanager.generalFilterWork(id1, id2, filter_name, argument, 
               filter_code, overlay, discard, did, vid);

    } catch (Exception e) {
      engine.getGUI().message("Client filter receiving error.", true);
    }
  }

  // receive raw iso data
  private void doCACHE_TRANSFER() {
System.out.println("Receiving raw iso data from the server.");
    engine.getGUI().message("Receiving raw iso data from the server.", false);

    try {
      // read the data set id
      long did = in.readLong();

      File file = new File(engine.getCacheDir(), String.valueOf(did));
      DataOutputStream outfile = 
                    new DataOutputStream(new BufferedOutputStream(new 
                                        FileOutputStream(file)));

      int option = in.readInt();

      if (option == 0) {

        int timesteps = in.readInt();
        outfile.writeInt(timesteps);

        for (int i=0; i<timesteps; i++) {
          outfile.writeFloat(in.readFloat());
          outfile.writeInt(in.readInt());
          outfile.writeInt(in.readInt());
          outfile.writeInt(in.readInt());
          int sz = in.readInt();

          for (int j=0; j<sz; j++) 
            outfile.writeFloat(in.readFloat());
        }      
      } else {
        int sz = in.readInt();
        byte[] arr = new byte[sz];
        in.readFully(arr);
        outfile.write(arr, 0, arr.length);
      }

      outfile.close();
    } catch (Exception e) {
       System.out.println("3D data cache writing error.");
    }
  }


} // end of class 



