package sv.engine;

import java.io.*;
import java.net.*;

/**
 * This class is a heart of the client for the collaboration and
 * connection to the server.
 */

public class cSendChannel {

  private InetAddress shost;   // server host address
  private int sport;           // server port number

  private int collab_opt;

  // Server state checker.
  // If the server does not responding, collaboration is meaningless.
  // It checks the server state and return result as a boolean value.
  private boolean isOk = true;

  // Is this client is master or slave
  private boolean isMaster = true;

  static SV2Engine engine;

  /**
   * SV2 collaborator
   * @param eng SV2Engine class
   * @param h server host name
   * @param p server port number
   * @param cp client port number
   * @param opt collaboration option
   */
  public cSendChannel(SV2Engine eng, String h, int p, int cp, int opt) {
    engine = eng;
    sport = p;
    collab_opt = opt;

    try {
      if (h == null) shost = InetAddress.getLocalHost();
      else           shost = InetAddress.getByName(h);

      initialConnect(cp);

    } catch (Exception e) {
      isOk = false;
      engine.getGUI().message("Can't connect the SV2 server, limited mode.", true);
    }
  }

  /**
   * set this client new master
   */
  public void setMaster(boolean b) {
    isMaster = b; 
  }

  /**
   * return if this client is master or slave
   */
  public boolean getIsMaster() {
    return isMaster;
  }

  /**
   * send exit information to the server
   * @param id client identification
   */
  public void notifyExit(long id) throws IOException {
    Socket sock = null;
    DataOutputStream out = null;

    if (isOk) {
      if (collab_opt == SV2Engine.COLLAB_TANGO) {
        // SOME ADDTIONAL WORK HERE
      }

      sock = new Socket(shost, sport);
      out = new DataOutputStream(sock.getOutputStream());

      out.writeInt(SV2Protocol.CLIENT_EXIT);
      out.writeLong(id);

      out.close();
      sock.close();
    }
  }

  /**
   * Send command to the server. The server will do somework 
   * after parsing the command . The server does not routing the command
   * to the clients.
   * @see broadcastCommand
   */ 
  public void serverCommand(byte[] datum) throws IOException {
    if (isMaster && isOk) {
      Socket sock = new Socket(shost, sport);
      DataOutputStream out = new DataOutputStream(sock.getOutputStream());

      out.writeInt(SV2Protocol.SERVER_DATA_WORK);
      out.writeInt(datum.length);
      out.write(datum, 0, datum.length);

      out.close();
      sock.close();
    }
  }

  /**
   * Send finter to the server. 
   * @see broadcastCommand
   */ 
  public void serverFilterAction(long id, String filter_name, 
                                 byte[] filter_code, String argument,
                                 long other_id, String mode1, String mode2)
                                            throws IOException {
    if (isMaster && isOk) {
      Socket sock = new Socket(shost, sport);
      DataOutputStream out = new DataOutputStream(sock.getOutputStream());

      out.writeInt(SV2Protocol.SERVER_FILTER_WORK);
      out.writeLong(id);
      out.writeUTF(filter_name);
      out.writeInt(filter_code.length);
      out.write(filter_code, 0, filter_code.length);
      out.writeUTF(argument);
      out.writeLong(other_id);
      out.writeUTF(mode1);
      out.writeUTF(mode2);

      out.close();
      sock.close();
    }
  }

  /**
   * Send command to the server no matter what the client is master or not.
   * @see serverCommand
   */ 
  public void forceServerCommand(byte[] datum) throws IOException {
    if (isOk) {
      Socket sock = new Socket(shost, sport);
      DataOutputStream out = new DataOutputStream(sock.getOutputStream());

      out.writeInt(SV2Protocol.SERVER_DATA_WORK);
      out.writeInt(datum.length);
      out.write(datum, 0, datum.length);

      out.close();
      sock.close();
    }
  }

  /**
   * Broadcast the command to the collaborator. The collaboration server
   * will just routing the command without doing any other work since
   * the collaboration server may not the SV2 server(in case of Tango).
   * This method is the only interface between the collaboration server
   * and the clients.
   * @see serverCommand
   */
  public void broadcastCommand(byte[] datum) throws IOException {
    if (isMaster && isOk) {
      if (collab_opt == SV2Engine.COLLAB_BUILTIN) {
        Socket sock = new Socket(shost, sport);
        DataOutputStream out = new DataOutputStream(sock.getOutputStream());

        out.writeInt(SV2Protocol.BROADCAST_COMMAND);
        out.writeInt(datum.length);
        out.write(datum, 0, datum.length);

        out.close();
        sock.close();
      } else if (collab_opt == SV2Engine.COLLAB_TANGO) {

      }
    }
  }
 
  // register this client to the server
  private void initialConnect(int cport) throws IOException {
    Socket sock = null;
    DataOutputStream out = null;
    DataInputStream in = null;

    sock = new Socket(shost, sport);

    out = new DataOutputStream(sock.getOutputStream());
    in = new DataInputStream(new 
                        BufferedInputStream(sock.getInputStream()));

    out.writeInt(SV2Protocol.REGISTER_CLIENT);
    out.writeInt(cport);
    out.writeInt(collab_opt);
    long id = in.readLong();
    isMaster = in.readBoolean();

    out.close();
    in.close();
    sock.close();

    if (id > 0) {
      if (isMaster) 
        engine.getGUI().message("Server " + 
                   shost.getHostName() + " ready. Master mode.", false);
      else
        engine.getGUI().message("Server " + 
                   shost.getHostName() + " ready. Slave mode.", false);
      engine.setID(id);
    } else {
      engine.getGUI().message("Server refused the connection, limited mode", 
                                                                      true);
      isOk = false;
    }
  }

  /**
   * is it a collaboration mode?
   */
  public boolean isCollabarable() {
    return isOk;
  }

} // end of class

