package sv.server;

import java.net.*;
import java.io.*;
import java.util.*;
import sv.data.*;
import sv.engine.SV2Protocol;
import sv.engine.SV2Action;
import sv.server.filter.VolumeFilter;

/** 
 * RecvHandler
 * work after getting connection from client or simulation engine
 */

public class RecvHandler extends Thread {

  boolean DEBUG = true;

  private Socket sock;
  private DataOutputStream out;
  private DataInputStream in;

  private SV2Server server;

  // cache directory
  File cache_dir = null;

  /**
   * constructor
   * @param s server socket
   * @param serv parent module(SV2Server)
   */
  public RecvHandler(Socket s, File cache, SV2Server serv) {
    sock = s;
    cache_dir = cache;
    server = serv;
  }

  /**
   * close the connection
   */
  public void closeConnection() throws IOException {
    in.close();
    out.close();
    sock.close();
  }

  /**
   * read commands from the client and do work
   */
  public void run() {
    int command = -1; 

    try {
      out = new DataOutputStream(sock.getOutputStream());
      in  = new DataInputStream(new 
                       BufferedInputStream(sock.getInputStream()));

      command = in.readInt();

      /////////////////////////////////////////////////////////////////
      // Now we have to decide what commands and data are received 
      // from the client. Normally, after receiving data, connection will
      // be disconnected. 

      doWork(command);                 // do the server work
 
      out.close();
      in.close();
      sock.close();                    // close the socket

    } catch (Exception e) {
      System.out.println("Server connection error: " + command);
      if (DEBUG) e.printStackTrace();

      try { out.close(); } catch (Exception E1) {}
      try { in.close(); } catch (Exception E2) {}
      try { sock.close(); } catch (Exception E3) {}
    }
  }

  // real worker
  private synchronized void doWork(int command) throws IOException {
    switch(command) {
      case SV2Protocol.KILL_SERVER:
        server.killServer();
        break;

      case SV2Protocol.REGISTER_CLIENT:
        doREGISTER_CLIENT();
        break;

      case SV2Protocol.CLIENT_EXIT:
        doCLIENT_EXIT();
        break;

      case SV2Protocol.SERVER_DATA_WORK:
        doSERVER_DATA_WORK();
        break;

      case SV2Protocol.BROADCAST_COMMAND:
        int sz = in.readInt();
        byte[] datum = new byte[sz];
        in.readFully(datum);
        server.getSender().broadcastDatum(datum);        
        break;

      case SV2Protocol.RAW_DATA_FEED:
        doRAW_DATA_FEED();
        break;
     
      case SV2Protocol.DEN_DATA_FEED:
        doDEN_DATA_FEED();
        break;

      case SV2Protocol.SERVER_FILTER_WORK:
        doSERVER_FILTER_WORK();
        break;

      default:
         System.err.println("Got undefined command: " + command);
     }
  }

  // register the client
  private void doREGISTER_CLIENT() throws IOException {
    int port = in.readInt();   
    int opt = in.readInt();
    InetAddress host = sock.getInetAddress();

    boolean isMaster = true;
    if (server.getSender().getNoOfCollaborator() > 0) 
      isMaster = false; 

    long id = server.getSender().addCollaborator(host, port, opt, isMaster);
    out.writeLong(id);
    out.writeBoolean(isMaster);
  }

  // client exits
  private void doCLIENT_EXIT() throws IOException {
    long id = in.readLong();
    server.getSender().removeCollaborator(id);
  }

  // do server filter work
  private void doSERVER_FILTER_WORK() throws IOException {
    long id1, id2;
    String filter_name, argument;
    byte[] filter_code;
    boolean overlay, discard;

    id1 = in.readLong();
    filter_name = in.readUTF();
    int sz = in.readInt();
    filter_code = new byte[sz];
    in.readFully(filter_code);
    argument  = in.readUTF();
    id2 = in.readLong();
    overlay = Boolean.valueOf(in.readUTF()).booleanValue();
    discard = Boolean.valueOf(in.readUTF()).booleanValue();
    server.getSender().generalFilterWork(id1, id2, filter_name, argument, 
               filter_code, overlay, discard);
  }

  // do server work
  private void doSERVER_DATA_WORK() throws IOException {
    int sz = in.readInt();
    byte[] datum = new byte[sz];
    in.readFully(datum);
    StringTokenizer st = new StringTokenizer(new String(datum), ";");
    
    // The first token is a command
    int comm = Integer.parseInt(st.nextToken());

    // next one is the data set id
    long id = Long.parseLong(st.nextToken());
    DataSetBrief brief = server.getSender().getDataSet(id);
    if (brief == null) 
      System.out.println("Server Error: Not found data set.");

    switch (comm) {
      case SV2Action.DELETE:
      case SV2Action.CLOSE:
        server.getSender().deleteDataSet(brief);
        break;

      case SV2Action.CUT:
        server.getSender().cutAction(brief);
        break;

      case SV2Action.COPY:
        server.getSender().copyAction(brief);
        break;

      case SV2Action.PASTE:
        server.getSender().pasteAction(brief);
        break;

      case SV2Action.VISUALIZE:
        if (brief.getRank() == TimeData.THREED) {
          int opt = Integer.parseInt(st.nextToken());
          if (opt == TimeData3D.ISOSURFACE_RENDERING) 
            server.getSender().marchingCube(brief,
                                 Long.parseLong(st.nextToken()), 
                                 Long.parseLong(st.nextToken()), 
                                 Byte.parseByte(st.nextToken()));

          else if (opt == TimeData3D.VOLUME_RENDERING) 
            server.getSender().volumeCompute(brief,
                                 Long.parseLong(st.nextToken()), 
                                 Long.parseLong(st.nextToken()), 
				 Double.valueOf(st.nextToken()).doubleValue(),
                                 Double.valueOf(st.nextToken()).doubleValue(),
                                 Double.valueOf(st.nextToken()).doubleValue(),
                                 Double.valueOf(st.nextToken()).doubleValue(),
                                 Double.valueOf(st.nextToken()).doubleValue(),
                                 Byte.parseByte(st.nextToken()));

        } else 
          server.getSender().sendDataSet(brief,
                                       Long.parseLong(st.nextToken()), 
                                       Long.parseLong(st.nextToken()), 
                                       Byte.parseByte(st.nextToken()));
        break;

      case SV2Action.ISOVALUE_CHANGE:
        server.getSender().marchingCube(brief,
                                 Float.valueOf(st.nextToken()).floatValue(),
                                 Long.parseLong(st.nextToken()), 
                                 -1, (byte)1);
        break;

      case SV2Action.VOLVALUE_CHANGE:
        server.getSender().volumeCompute(brief,
                                 Long.parseLong(st.nextToken()), 
                                 -1,
                                 Double.valueOf(st.nextToken()).doubleValue(),
                                 Double.valueOf(st.nextToken()).doubleValue(),
                                 Double.valueOf(st.nextToken()).doubleValue(),
                                 Double.valueOf(st.nextToken()).doubleValue(),
                                 Double.valueOf(st.nextToken()).doubleValue(),
                                 (byte)1);
         break;
 
      case SV2Action.VOLBULK_REQUEST:
        long bulk_vid = Long.parseLong(st.nextToken()); 
        int bulk_mode = Integer.parseInt(st.nextToken());
	double requested_value = Double.valueOf(st.nextToken()).doubleValue();
        int bulk_sz = Integer.parseInt(st.nextToken());
        double bulk_vals[] = new double[bulk_sz];
        for (int i=0; i<bulk_sz; i++) 
          bulk_vals[i] = Double.valueOf(st.nextToken()).doubleValue();

        server.getSender().volumeBulkCompute(brief, bulk_vid, bulk_mode, requested_value, bulk_vals);

        break;

      case SV2Action.DOWNSIZE:
      case SV2Action.DOWNSIZE_SMOOTHE:
        int factor = Integer.parseInt(st.nextToken());
        server.getSender().r3BasicFilter(comm, brief, factor);
        break;

      case SV2Action.SMOOTHE:
        server.getSender().r3BasicFilter(comm, brief, 0);
        break;

      default:
        System.out.println("Unknown server's data action command.");
        return;
    } 
  }

  // receive raw data set
  private void doRAW_DATA_FEED() throws IOException {

    // read title 
    int sz = in.readInt();
    byte[] byteArr = new byte[sz];
    in.read(byteArr, 0, sz);
    String title = new String(byteArr);

    // read the number of time steps
    int noOfTimeData = in.readInt();
    TimeData[] readed = new TimeData[noOfTimeData];

    for (int k=0; k<noOfTimeData; k++) {
      int type = in.readInt();

      switch (type) {
        case 101:    // point data
          readed[k] = DataReader.parse101(in);
          break;
        case 102:   
          readed[k] = DataReader.parse102(in);
          break;
        case 103:   
          readed[k] = DataReader.parse103(in);
          break;
        case 104:   
          readed[k] = DataReader.parse104(in);
          break;
        case 105:   
          readed[k] = DataReader.parse105(in);
          break;
        case 201:    // 2d data
          readed[k] = DataReader.parse201(in);
          break;
        case 301:    // 3d data
          readed[k] = DataReader.parse301(in);
          break;
        default:
          System.out.println("Unknown data type error.");
          return;
      }
    }
 
    DataSet dataSet = new DataSet(readed);
    dataSet.setDataName(title);
    server.getSender().receivedData(dataSet);
  }

  // receive density data
  private void doDEN_DATA_FEED() throws IOException {

    // read title 
    int sz = in.readInt();
    byte[] byteArr = new byte[sz];
    in.read(byteArr, 0, sz);
    String title = new String(byteArr);

    // read the data information
    int xsize = in.readInt();
    int ysize = in.readInt();
    int zsize = in.readInt();
    int totsz = in.readInt();

    byte[] arr = new byte[totsz];
    in.readFully(arr);

    String id = String.valueOf(System.currentTimeMillis());
    String path = cache_dir.getAbsolutePath();
    if (!path.endsWith(File.separator))
       path = path + File.separator + id;

    File file = new File(cache_dir, id);
    DataOutputStream outfile = 
 new DataOutputStream(new BufferedOutputStream(new FileOutputStream(file)));

    outfile.writeInt(1);   // the number of time step

    float min = Float.MAX_VALUE;
    float max = Float.MIN_VALUE;

    outfile.writeFloat(0.0f);   // time value
    outfile.writeInt(xsize);
    outfile.writeInt(ysize);
    outfile.writeInt(zsize);

    for (int j=0; j<totsz; j++) { 
      float val;
      if (arr[j] < 0) val = (float)arr[j] + 256f;
      else            val = (float)arr[j];
      outfile.writeFloat(val);

      if (val < min) min = val;
      if (val > max) max = val;
    }

    outfile.close();

    TimeData3D td = new TimeData3D(xsize, ysize, zsize, min, max, id);
    DataSet dataSet = new DataSet(td);
    dataSet.setDataName(title);
    server.getSender().receivedData(dataSet);
  }
} // end of class 

