package sv.server;

import java.io.*;
import java.net.*;
import java.util.*;
import sv.data.*;
import sv.engine.SV2Protocol;
import sv.engine.SV2Action;
import sv.server.filter.*;
import sv.util.*;

public class SendChannel {

  boolean DEBUG = true;

  // collaborator vector
  Vector collabVector;

  // data set vector in the server
  Vector dataVector;

  // cache directory
  File cache_dir = null;

  // clipboard data set
  private DataSetBrief clipboard = null;

  /**
   * constructor
   * @param cache cache directory
   */
  public SendChannel(File cache) {
    cache_dir = cache;
    collabVector = new Vector();
    dataVector = new Vector();
  }

  /**
   * return data from the data set id
   */
  public DataSetBrief getDataSet(long did) {
    for (Enumeration e=dataVector.elements(); e.hasMoreElements();) {
      DataSetBrief brief = (DataSetBrief)e.nextElement();
      if (brief.getID() == did) 
        return brief;
    }
    return null;
  }

  /**
   * received data set from the simulator or file reader.
   * broadcast received data information to the clients
   */
  public void receivedData(DataSet data) {
    if (collabVector.size() < 1) {
      data.garbageCollect();
      data = null;
      System.out.println("There is no clients. Dataset discarded.");
      return;
    }
  
    for (Enumeration e=dataVector.elements(); e.hasMoreElements();) {
      DataSetBrief brief = (DataSetBrief)e.nextElement();
      DataSet[] dataSet = brief.getDataSet();

      for (int i=0; i<dataSet.length; i++) 
        if (dataSet[i].getDataName().equals(data.getDataName())) {
          if (brief.getViewerID() > 0) {
             System.out.println("Data was already visualized! ");
             System.out.println("Received data won't forward to client(s) in this version.");
          }
          
          int sz = data.getNoOfTimes();
          for (int j=0; j<sz; j++)
            dataSet[i].addTimeData(data.getTimeData(j));

          data.garbageCollect();
          data = null;
          return;
        }
    }

    DataSetBrief brief = new DataSetBrief(data.getDataName(), 
                                          data.getNoOfTimes(), 
                                          data.getTimeData(0).getRank(), 
                                          data.getDataSize());
    brief.setID(System.currentTimeMillis());      
    brief.addDataSet(data);
    dataVector.addElement(brief);

    if (brief.getRank() == TimeData.THREED) {
      System.out.print("Precomputing for the volume data...");
      String path = cache_dir.getAbsolutePath();
      if (!path.endsWith(File.separator))
        path = path + File.separator + brief.getID();
      TimeData3D td = (TimeData3D)data.getTimeData(0);
      VolumeFilter.precompute(path, td.getGridData(cache_dir), 
       td.getzmin(), td.getzmax(), td.getDimX(), td.getDimY(), td.getDimZ());
      System.out.println(" done");
    }

    Socket sock = null;
    DataOutputStream out = null;

    for (Enumeration e=collabVector.elements(); e.hasMoreElements();) {
      Collaborator unit = (Collaborator)e.nextElement();

      try {
        sock = new Socket(unit.host_addr, unit.port_number);
        out = new DataOutputStream(sock.getOutputStream());
        
        out.writeInt(SV2Protocol.HEAD_DATA_FEED);
        out.writeUTF(brief.getDataName());
        out.writeInt(brief.getNoOfTime());
        out.writeInt(brief.getRank());
        out.writeLong(brief.getDataSize());
        out.writeLong(brief.getID());

        out.close();
        sock.close();

      } catch (Exception E) {
        System.out.println(unit.host_addr.getHostName() + 
          " does not responding. Remove this client.");
        collabVector.removeElement(unit);
      }
    }
  }

  /**
   * basic filter for large 3d data set
   * @param comm filter command
   * @param brief data set
   * @param factor argument
   */
  public void r3BasicFilter(int comm, DataSetBrief brief, int factor) {
System.out.println("r3BasicFilter: "+comm+" "+brief.getDataName()+" "+ factor);  

    if (brief == null) return;

    DataSet[] dss = brief.getDataSet();

    // ASSUME ONLY FOR ONE DATA SET AND ONE TIME DATA
    int sz = dss[0].getNoOfTimes();
    String str = "";
    TimeData res = null;

    TimeData3D timeData = (TimeData3D)dss[0].getTimeData(0);

    if (comm == SV2Action.DOWNSIZE) {
      if (factor < 2) {
        System.out.println("Meaningless downsizing.");
        return;
      }
      res = timeData.downsize(cache_dir, factor);
      str = "_d" + String.valueOf(factor);
      

    } else if (comm == SV2Action.SMOOTHE) {
      res = timeData.smoothe(cache_dir);
      str = "_s";

    } else if (comm == SV2Action.DOWNSIZE_SMOOTHE) {
      if (factor < 2) {
        System.out.println("Meaningless downsizing.");
        return;
      }
      res = timeData.downsize_smoothe(cache_dir, factor);
      str = "_ds" + String.valueOf(factor);
    }

    DataSet dataSet = new DataSet(res);
    dataSet.setDataName(dss[0].getDataName() + str);
    receivedData(dataSet);
  }

  /**
   * cut the dataset
   * @param brief dataset 
   */
  public void cutAction(DataSetBrief brief) {
    copyAction(brief);
    deleteDataSet(brief);
  }

  /**
   * copy the dataset
   * @param brief dataset 
   */
  public void copyAction(DataSetBrief brief) {
    clipboard = new DataSetBrief(brief.getDataName(), brief.getNoOfTime(),
                                 brief.getRank(), brief.getDataSize());

    clipboard.setID(brief.getID());
    clipboard.setViewerID(brief.getViewerID());
    clipboard.setPrivate(brief.isPrivate());
    DataSet[] ds = brief.getDataSet();
    if (ds != null) 
      for (int i=0; i<ds.length; i++)
        clipboard.addDataSet(ds[i].getClone());
  }

  /**
   * paste the dataset
   * @param brief dataset 
   */
  public void pasteAction(DataSetBrief brief) {
    DataSet[] ds = clipboard.getDataSet();
    if (ds != null) 
      for (int i=0; i<ds.length; i++)
        brief.addDataSet(ds[i]);
    brief.updateInfo();

    if (clipboard.getViewerID() < 0 && brief.getViewerID() > 0) {
System.out.println("Paste: adding data from the server.");

      for (int i=0; i<collabVector.size(); i++) {
        Collaborator unit = (Collaborator)collabVector.elementAt(i);
        sendCookedDataSet(unit, clipboard, brief.getViewerID());
      }
    }

    clipboard.setClear();
    clipboard = null;
  }

  /**
   * delete the dataset
   * @param brief dataset 
   */
  public void deleteDataSet(DataSetBrief brief) {
    if (brief == null) return;

    ////////////////////////////////////////////////////
    // IT SHOULD ALSO DELETE THE CACHED FILE
    ////////////////////////////////////////////////////
    if (DEBUG) System.out.println("Data set deletion: " + brief.getDataName());
    dataVector.removeElement(brief);
    brief.setClear();
  }

  /**
   * send the dataset to the client(s)
   * @param brief dataset 
   * @param vid viewer id
   * @param cid client id
   * @param MasterSlave master or slave
   */
  public void sendDataSet(DataSetBrief brief, long vid, long cid, byte MasterSlave) {

    if (brief == null) return;

    if (MasterSlave == 1) {              // master
      if (brief.getViewerID() > 0) {     // refuse (Protocol error)
        sendExceptionMsg(cid, SV2Protocol.PROTOCOL_ERROR);
      } else {                           // broadcast the data
        for (int i=0; i<collabVector.size(); i++) {
          Collaborator unit = (Collaborator)collabVector.elementAt(i);
          sendCookedDataSet(unit, brief, vid);
        }
      }
    } else {                             // slave
      if (brief.getViewerID() > 0) {     // send the data
        for (int i=0; i<collabVector.size(); i++) {
          Collaborator unit = (Collaborator)collabVector.elementAt(i);
          if (cid == unit.id) {
            sendCookedDataSet(unit, brief, vid);
            break;
          }
        }
      } else {                           // refuse (privilege not allows)
        sendExceptionMsg(cid, SV2Protocol.ACCESS_DENIED);
      }
    }
  }

  /**
   * filter processing function in the server
   * @param id1 first data set identification
   * @param id2 second data set identification, used only for merging
   * @param filter_name filter name
   * @param argument filter argument
   * @param filter_code filter source code
   * @param overlay overlay mode
   * @param discard discard mode
   */
  public void generalFilterWork(long id1, long id2, String filter_name, 
                                String argument, byte[] filter_code, 
                                boolean overlay, boolean discard) {

    System.out.println("Server filter action: " + filter_name);
    DataSetBrief brief = null;

    if (filter_code.length < 1) {
      System.out.println("Server filter action error(null).");
      return;
    } 
   
    try {
      FilterClassLoader loader = new FilterClassLoader(filter_code);
      Class rfilter = loader.loadClass(filter_name, true);
      if (rfilter == null) {
        System.out.println("Server filter action error(Format).");
        return;
      }

      Filter filter = (Filter)rfilter.newInstance();

      if (filter.getStyle() == Filter.NORMAL) {
        DataSetBrief dataSet = getDataSet(id1);

        DataSet[] ds = dataSet.getDataSet();
        int noOfInputs = filter.getNoOfArgs();

        for (int elt=0; elt<ds.length; elt++) {
          DataSet data = ds[elt];
          filter.clear();

          if (noOfInputs == 1) {
             float val = 0.0f;
             try {
                val = Float.valueOf(argument).floatValue();
             } catch (Exception ee) {
                System.out.println("Server filter action error(Input).");
                return;
             }

             filter.assignDataSet(data);
             filter.assignArgs(val);

          } else if (noOfInputs == 2) {
             float start = 0, end = 0;
             int where = argument.indexOf("-");
             if (where < 0) {
                System.out.println("Server filter action error(Usage).");
                return;
             } else {
                try {
                  start = 
                   Float.valueOf(argument.substring(0, where).trim()).floatValue();
                  end = 
                   Float.valueOf(argument.substring(where+1).trim()).floatValue();
                } catch (Exception ee) {
                    System.out.println("Server filter action error(Usage).");
                    return;
                }

                if (end < start) {
                  System.out.println("Server filter action error(Range).");
                  return;
                }
             }
             filter.assignDataSet(data);
             filter.assignArgs(start, end);

          } else         // if it doesn't need an argument
             filter.assignDataSet(data);

          DataSet newdata;
          try { 
            newdata = (DataSet)filter.performFilter();
  	  } catch (Exception ee) {
            System.out.println("Server filter action error(Code).");
            return;
          }

          if (newdata == null) {
            System.out.println("Server filter action error(null result).");
            return;
          }

          if (overlay) {
	    dataSet.addDataSet(newdata);

          } else {
            if (brief == null)
              brief = new DataSetBrief(newdata.getDataName(), 
                                     newdata.getNoOfTimes(), 
                                     filter.getOutDimension(),
                                     newdata.getDataSize());

            brief.addDataSet(newdata);
          }
        }

        if (brief != null) {
          brief.setID(System.currentTimeMillis());
          brief.updateInfo();
          brief.setViewerID(System.currentTimeMillis() + 1);
          dataVector.addElement(brief);
        }

	if (discard)
	  deleteDataSet(dataSet);

      } else if (filter.getStyle() == Filter.MULTIPLE) {

        DataSet[] dss1 = getDataSet(id1).getDataSet();
        DataSet[] dss2 = getDataSet(id2).getDataSet();

        if (dss1.length!=1 || dss2.length!=1) {
          System.out.println("Multi-data merging does not support yet.");
          return;
        }
     
        int noOfInputs = filter.getNoOfArgs();
        filter.clear();

        if (noOfInputs >= 1) {
          System.out.println("Merging with arguments does not support yet.");
        } else {        // if it doesn't need an argument
          filter.assignDataSet(dss1[0]);
          filter.assignDataSet(dss2[0]);
        }

        DataSet newdata;
        try { 
          newdata = (DataSet)filter.performFilter();
        } catch (Exception ee) {
          System.out.println("Server filter action error(Code).");
          return;
        }

        if (newdata == null) {
          System.out.println("Server filter action error(null result).");
          return;
        }

        if (overlay) {
          getDataSet(id1).addDataSet(newdata);

        } else {
          if (brief == null)
            brief = new DataSetBrief(newdata.getDataName(), 
                                   newdata.getNoOfTimes(), 
                                   filter.getOutDimension(),
                                   newdata.getDataSize());

          brief.addDataSet(newdata);
        }

        if (brief != null) {
          brief.setID(System.currentTimeMillis());
          brief.updateInfo();
          brief.setViewerID(System.currentTimeMillis() + 1);
          dataVector.addElement(brief);
        }

        if (discard) 
          System.out.println("Discarding original data for merging does not support yet.");

      }

    } catch (Exception e) {
e.printStackTrace();
      System.out.println("Server filter action error(Conversion).");
    }


    // NOW IT SHOULD BROADCAST THE RESULT
    Socket sock = null;
    DataOutputStream out = null;

    for (Enumeration e=collabVector.elements(); e.hasMoreElements();) {
      Collaborator unit = (Collaborator)e.nextElement();

      try {
        sock = new Socket(unit.host_addr, unit.port_number);
        out = new DataOutputStream(sock.getOutputStream());
        
        out.writeInt(SV2Protocol.CLIENT_FILTER_WORK);
        out.writeLong(id1);
        out.writeUTF(filter_name);
        out.writeInt(filter_code.length);
        out.write(filter_code, 0, filter_code.length);
        out.writeUTF(argument);
        out.writeLong(id2);
        out.writeUTF(String.valueOf(overlay));
        out.writeUTF(String.valueOf(discard));

        if (brief == null) {
          out.writeInt(0);
        } else {
          out.writeInt(1);
          out.writeLong(brief.getID());
          out.writeLong(brief.getViewerID());
        }

        out.close();
        sock.close();

      } catch (Exception E) {
        System.out.println(unit.host_addr.getHostName() + 
          " does not responding. Remove this client.");
        collabVector.removeElement(unit);
      }
    }
  }

  /**
   * marching cube without threshold (find middle value)
   * @param brief dataset 
   * @param vid viewer id
   * @param cid client id
   * @param MasterSlave master or slave
   */
  public void marchingCube(DataSetBrief brief, long vid, long cid, byte MasterSlave) {
    DataSet[] dataSets = brief.getDataSet();
    DataSet dataSet = dataSets[0];
    TimeData3D timedata = (TimeData3D)dataSet.getTimeData(0);
    float min = timedata.getzmin();
    float max = timedata.getzmax();
    float thresh = (min + max) / 2.0f;

    marchingCube(brief, thresh, vid, cid, MasterSlave);

    // SEND THE RAW DATA ALSO FOR SIMPLICITY AT THE FIRST TIME
    if (MasterSlave == 1) {              // master
      for (int i=0; i<collabVector.size(); i++) {
        Collaborator unit = (Collaborator)collabVector.elementAt(i);
        sendRawIsoDataSet(unit, brief);
      }
    } else {                             // slave
      if (brief.getViewerID() > 0) {     // send the data
        for (int i=0; i<collabVector.size(); i++) {
          Collaborator unit = (Collaborator)collabVector.elementAt(i);
          if (cid == unit.id) {
            sendRawIsoDataSet(unit, brief);
            break;
          }
        }
      } 
    }
  }


  /**
   * send the isosurface data to the client(s)
   * @param brief dataset 
   * @param thresh threshold
   * @param vid viewer id
   * @param cid client id
   * @param MasterSlave master or slave
   */
  public void marchingCube(DataSetBrief brief, float thresh, long vid, long cid, byte MasterSlave) {

    if (brief == null) return;

    DataSet[] dataSets = brief.getDataSet();
    for (int i=0; i<dataSets.length; i++) {
      int sz = dataSets[i].getNoOfTimes();
      for (int j=0; j<sz; j++) {
        TimeData3D timeData = (TimeData3D)dataSets[i].getTimeData(j);
        timeData.marchingCube(cache_dir, thresh);         
      }
    }
  
    if (MasterSlave == 1) {              // master
      for (int i=0; i<collabVector.size(); i++) {
        Collaborator unit = (Collaborator)collabVector.elementAt(i);
        sendIsoDataSet(unit, brief, vid);
      }
    } else {                             // slave
      if (brief.getViewerID() > 0) {     // send the data
        for (int i=0; i<collabVector.size(); i++) {
          Collaborator unit = (Collaborator)collabVector.elementAt(i);
          if (cid == unit.id) {
            sendIsoDataSet(unit, brief, vid);
            break;
          }
        }
      } else {                           // refuse (privilege not allows)
        sendExceptionMsg(cid, SV2Protocol.ACCESS_DENIED);
      }
    }
  }

  /**
   * send the volume image data to the client(s)
   * @param brief dataset 
   * @param vid viewer id
   * @param cid client id
   * @param red red component od the ray color
   * @param green green component od the ray color
   * @param blue blue component od the ray color
   * @param rotx x rotation value
   * @param roty y rotation value
   * @param MasterSlave master or slave
   */
  public void volumeCompute(DataSetBrief brief, long vid, long cid, 
                double red, double green, double blue,
                double rotx, double roty, byte MasterSlave) {

    if (brief == null) return;

    // ONLY ALLOW ONE TIME STEPS
    DataSet[] dataSet = brief.getDataSet();
    TimeData3D timeData = (TimeData3D)dataSet[0].getTimeData(0);

    double[] vars = new double[5];
    vars[0] = red;
    vars[1] = green;
    vars[2] = blue;
    vars[3] = rotx;
    vars[4] = roty;
    VolumeFilter.setVariables(vars);

    byte[] data_arr = new byte[3*256*256];
    String path = cache_dir.getAbsolutePath();
    if (!path.endsWith(File.separator))
      path = path + File.separator + brief.getID();

    int result = VolumeFilter.executeFilter(path, data_arr);

    if (result == 0) {

      if (MasterSlave == 1) {              // master
        for (int i=0; i<collabVector.size(); i++) {
          Collaborator unit = (Collaborator)collabVector.elementAt(i);
          sendVolumeImage(unit, brief, vid, data_arr);
        }
      } else {                             // slave
        if (brief.getViewerID() > 0) {     // send the data
          for (int i=0; i<collabVector.size(); i++) {
            Collaborator unit = (Collaborator)collabVector.elementAt(i);
            if (cid == unit.id) {
              sendVolumeImage(unit, brief, vid, data_arr);
              break;
            }
          }
        } else {                           // refuse (privilege not allows)
          sendExceptionMsg(cid, SV2Protocol.ACCESS_DENIED);
        }
      }
    } else 
      System.out.println("Volume image calculation error.");
  }

  /**
   * send the bulk volume image data to the client(s)
   * @param brief dataset 
   * @param vid viewer id
   * @param mode x or y rotation mode
   * @param requested rotation value
   * @param vals rotation values
   */
  public void volumeBulkCompute(DataSetBrief brief, long vid, int mode, 
                double requested, double[]  vals) {

    if (brief == null) return;

    // ONLY ALLOW ONE TIME STEPS
    DataSet[] dataSet = brief.getDataSet();
    TimeData3D timeData = (TimeData3D)dataSet[0].getTimeData(0);
    
    double[] vars = VolumeFilter.getVariables();
    byte[][] data_arrs = new byte[vals.length][3*256*256];

    String path = cache_dir.getAbsolutePath();
    if (!path.endsWith(File.separator))
      path = path + File.separator + brief.getID();

    for (int i=0; i<vals.length; i++) {    
      if (mode == 0) {   // x rotation
        vars[3] = vals[i];
        vars[4] = requested;
      } else {
        vars[3] = requested;
        vars[4] = vals[i];
      }
      VolumeFilter.setVariables(vars);

      byte[] data_arr = new byte[3*256*256];

      int result = VolumeFilter.executeFilter(path, data_arr);

System.out.println("Volume image computing: " + i + " out of " + vals.length);

      if (result == 0) 
        data_arrs[i] = data_arr;
      else {
        System.out.println("Volume computation error...");
        return;
      }
    }
     
    // assume master
    for (int i=0; i<collabVector.size(); i++) {
       Collaborator unit = (Collaborator)collabVector.elementAt(i);
       Socket sock = null;
       DataOutputStream out = null;

       brief.setViewerID(vid);

       try {
         sock = new Socket(unit.host_addr, unit.port_number);
         out = new DataOutputStream(sock.getOutputStream());

         // first send command
         out.writeInt(SV2Protocol.VOLUME_IMAGE_BULK);
 
         // send the viewer id
         out.writeLong(vid);

         // send the data id
         out.writeLong(brief.getID());
 
         out.writeInt(vals.length);
         for (int j=0; j<vals.length; j++) { 
           double[] old_vars = new double[vars.length];
           System.arraycopy(vars, 0, old_vars, 0, vars.length);
           if (mode == 0)     // x rotation
             old_vars[3] = vals[j];
           else
             old_vars[4] = vals[j];

           out.writeInt(old_vars.length);
           for (int k=0; k<old_vars.length; k++) 
             out.writeDouble(old_vars[k]);

           out.write(data_arrs[j], 0, data_arrs[j].length);
         }

         out.close();
         sock.close();

       } catch (Exception e) {
         if (DEBUG) e.printStackTrace();
       }

    }
  }

  /**
   * just broadcast received datum from the master client to each slave 
   * In other than builtin collaboration mode, it will be never used.
   */
  public void broadcastDatum(byte[] datum) {

    Socket sock = null;
    DataOutputStream out = null;

    for (Enumeration e=collabVector.elements(); e.hasMoreElements();) {
      Collaborator unit = (Collaborator)e.nextElement();
      if (unit.isMaster) continue;

      try {
        sock = new Socket(unit.host_addr, unit.port_number);
        out = new DataOutputStream(sock.getOutputStream());
        
        out.writeInt(SV2Protocol.BROADCAST_COMMAND);
        out.writeInt(datum.length);
        out.write(datum, 0, datum.length);

        out.close();
        sock.close();

      } catch (Exception E) {
        System.out.println(unit.host_addr.getHostName() + 
          " does not responding. Remove this client.");
        collabVector.removeElement(unit);
      }
    }
  }

  /**
   * return the number of collaborator
   */
  public int getNoOfCollaborator() {
    return collabVector.size();
  }

  /**
   * add collaborator
   * @param h collaborator host address 
   * @param p collaborator socket port
   * @param opt collaboration option
   * @param b if master, b is true
   * @return collaborator identification number
   */
  public long addCollaborator(InetAddress h, int p, int opt, boolean b) {

    int sz = collabVector.size();
    boolean yesdo = true;
    Collaborator unit;

    if (sz > 1) {
      unit = (Collaborator)collabVector.firstElement();
      if (unit.collab_mode != opt)
        yesdo = false;
    }

    if (yesdo) {
      unit = new Collaborator(h, p, opt, b);
      collabVector.addElement(unit);
      if (DEBUG) {
        System.out.println("Add collaborator: " + h.getHostName());
      }
      return unit.id;
    } 

    return -1;
  }

  /**
   * remove collaborator
   * @param id collaborator's identification
   */
  public void removeCollaborator(long id) {
    boolean special = false;

    for (Enumeration e=collabVector.elements(); e.hasMoreElements();) {
      Collaborator unit = (Collaborator)e.nextElement();
      if (id == unit.id) {
        if (unit.isMaster)
          special = true;        
        
        if (DEBUG) {
          System.out.println("Remove Collaborator: " + unit.host_addr.getHostName());
        }

        collabVector.removeElement(unit);
        unit = null;
        break;
      }
    }
    
    if (special) {         // elect new master
      //////////////////////////////////////////////////
      // IT SHOULD CHECK CLIENT'S COMPUTING POWER
      // WARNING!! how about fail all clients are refused to be master

      for (Enumeration e=collabVector.elements(); e.hasMoreElements();) {
        Collaborator unit = (Collaborator)e.nextElement();
        if (electMaster(unit)) {
          if (DEBUG) {
            System.out.println("Master election: " + unit.host_addr.getHostName());
          }
          break;
        }
      }
    }
  }

  // elect the new master
  private boolean electMaster(Collaborator unit) {
    Socket sock = null;
    DataOutputStream out = null;

    try {
      sock = new Socket(unit.host_addr, unit.port_number);
      out = new DataOutputStream(sock.getOutputStream());

      out.writeInt(SV2Protocol.NEW_MASTER);

    } catch (Exception e) {
      return false;
    }

    try {
      out.close();
      sock.close();
    } catch (Exception e) { 
      return true;
    }
 
    return true;
  }


  // send exception message
  private void sendExceptionMsg(long cid, int msg) {
    Socket sock = null;
    DataOutputStream out = null;

    for (Enumeration e=collabVector.elements(); e.hasMoreElements();) {
      Collaborator unit = (Collaborator)e.nextElement();

      if (cid == unit.id) {
        try {
          sock = new Socket(unit.host_addr, unit.port_number);
          out = new DataOutputStream(sock.getOutputStream());

          out.writeInt(msg);
          
          out.close();
          sock.close();
          return;          
          
        } catch (Exception e2) {
          if (DEBUG) e2.printStackTrace();
          return;
        }
      }
    }     
  }

  // send the real data set to a client
  private void sendCookedDataSet(Collaborator unit, DataSetBrief brief, 
                                                              long vid) {
    Socket sock = null;
    DataOutputStream out = null;

    brief.setViewerID(vid);

    try {
      sock = new Socket(unit.host_addr, unit.port_number);
      out = new DataOutputStream(sock.getOutputStream());

      // first send command
      out.writeInt(SV2Protocol.COOKED_DATA_FEED);
 
      // send the viewer id
      out.writeLong(vid);

      // send the data id
      out.writeLong(brief.getID());

      // send the data set
      DataSet[] dataSets = brief.getDataSet();
      out.writeInt(dataSets.length);
      for (int i=0; i<dataSets.length; i++) {
        out.writeUTF(dataSets[i].getDataName());

        int sz = dataSets[i].getNoOfTimes();
        out.writeInt(sz);
        for (int j=0; j<sz; j++) {
          TimeData timeData = dataSets[i].getTimeData(j);
          timeData.sendingData(out);           /// TEMPORARY
        }
      }

      out.close();
      sock.close();

    } catch (Exception e) {
      if (DEBUG) e.printStackTrace();
    }
  }

  // send the raw isosurface data to a client
  private void sendRawIsoDataSet(Collaborator unit, DataSetBrief brief) {

    Socket sock = null;
    DataOutputStream out = null;

    try {
      sock = new Socket(unit.host_addr, unit.port_number);
      out = new DataOutputStream(sock.getOutputStream());

      // first send command
      out.writeInt(SV2Protocol.CACHE_TRANSFER);
 
      // send the data id
      out.writeLong(brief.getID());

      // send the first data set only
      DataSet[] dataSets = brief.getDataSet();
      int sz = dataSets[0].getNoOfTimes();
      TimeData3D td = (TimeData3D)dataSets[0].getTimeData(0);
  
      if (!td.isCached()) {
        out.writeInt(0);     // cache information

        out.writeInt(sz);

        for (int j=0; j<sz; j++) {
          TimeData3D timedata = (TimeData3D)dataSets[0].getTimeData(j);

          out.writeFloat(timedata.getTime().getAbsTime());

          out.writeInt(timedata.getDimX());
          out.writeInt(timedata.getDimY());
          out.writeInt(timedata.getDimZ());

          float[] grid = timedata.getGridData(cache_dir);
          out.writeInt(grid.length);
          for (int i=0; i<grid.length; i++) 
            out.writeFloat(grid[i]);
        }
      } else {
        DataInputStream in = new DataInputStream(new BufferedInputStream(
                   new FileInputStream(new File(cache_dir, td.getCacheName())))); 

        byte[] b = new byte[in.available()];
        in.readFully(b);
        in.close();

        out.writeInt(1);     // cache information
        out.writeInt(b.length);
        out.write(b, 0, b.length);
      }

      out.close();
      sock.close();

    } catch (Exception e) {
      if (DEBUG) e.printStackTrace();
    }
  }

  // send the isosurface data to a client
  private void sendIsoDataSet(Collaborator unit, DataSetBrief brief, 
                                               long vid) {
    Socket sock = null;
    DataOutputStream out = null;

    brief.setViewerID(vid);

    try {
      sock = new Socket(unit.host_addr, unit.port_number);
      out = new DataOutputStream(sock.getOutputStream());

      // first send command
      out.writeInt(SV2Protocol.COOKED_DATA_FEED);
 
      // send the viewer id
      out.writeLong(vid);

      // send the data id
      out.writeLong(brief.getID());

      // send the data set
      DataSet[] dataSets = brief.getDataSet();
      out.writeInt(dataSets.length);
      for (int i=0; i<dataSets.length; i++) {
        out.writeUTF(dataSets[i].getDataName());

        int sz = dataSets[i].getNoOfTimes();
        out.writeInt(sz);

        for (int j=0; j<sz; j++) {
          out.writeInt(302);    // magic number

          TimeData3D timedata = (TimeData3D)dataSets[i].getTimeData(j);
          out.writeFloat(timedata.getTime().getAbsTime());

          out.writeInt(timedata.getDimX());
          out.writeInt(timedata.getDimY());
          out.writeInt(timedata.getDimZ());
          out.writeFloat(timedata.getThreshold());
          out.writeFloat(timedata.getzmin());
          out.writeFloat(timedata.getzmax());

          String path = cache_dir.getAbsolutePath();
          if (!path.endsWith(File.separator))
            path = path + File.separator + 
                timedata.getCacheName() +"." + String.valueOf(timedata.getThreshold());

          DataInputStream in = new DataInputStream(new BufferedInputStream(
                  new FileInputStream(new File(cache_dir, 
                  timedata.getCacheName() +".iso"))));

          int num_corners = in.readInt();
          out.writeInt(num_corners);
System.out.println("Sending " + num_corners + " corner information.");
          for(int n=1; n<=num_corners; n++) {
            out.writeFloat(in.readFloat());
            out.writeFloat(in.readFloat());
            out.writeFloat(in.readFloat());
            out.writeFloat(in.readFloat());
            out.writeFloat(in.readFloat());
            out.writeFloat(in.readFloat());
          }

          int size = in.readInt();
          out.writeInt(size);
System.out.println("Sending " + size + " connection information");
          for (int n = 0; n<size; n++) {
            out.writeInt(in.readInt());
            out.writeInt(in.readInt());
            out.writeInt(in.readInt());
          }     

          in.close();
        }
      }
      out.close();
      sock.close();

    } catch (Exception e) {
      if (DEBUG) e.printStackTrace();
    }
  }

  // send the volume image data to a client
  private void sendVolumeImage(Collaborator unit, DataSetBrief brief, 
                                                     long vid, byte[] arr) {
    Socket sock = null;
    DataOutputStream out = null;

    brief.setViewerID(vid);

    try {
      sock = new Socket(unit.host_addr, unit.port_number);
      out = new DataOutputStream(sock.getOutputStream());

      // first send command
      out.writeInt(SV2Protocol.VOLUME_IMAGE_FEED);
 
      // send the viewer id
      out.writeLong(vid);

      // send the data id
      out.writeLong(brief.getID());

      double[] vars = VolumeFilter.getVariables();
      out.writeInt(vars.length);
      for (int i=0; i<vars.length; i++) 
        out.writeDouble(vars[i]);

      out.write(arr, 0, arr.length);

      out.close();
      sock.close();

    } catch (Exception e) {
      if (DEBUG) e.printStackTrace();
    }
  }

  // definition class of collaborator information 
  class Collaborator {
    long id;
    InetAddress host_addr;
    int port_number;
    int collab_mode;
    boolean isMaster;

    // the lower, the better (optional)
    long machine_power = 0;
    long comm_power = 0;
    long overall_level = 0;    // 0 - 7

    Collaborator(InetAddress h, int p, int opt, boolean b) {
      id = System.currentTimeMillis();
      host_addr = h;
      port_number = p;
      collab_mode = opt;     
      isMaster = b;
    }
  }

} // end of class

