package sv.util;

import java.util.*;
import java.io.*;
import java.awt.image.*;

public class PPMImgProducer implements ImageProducer {

  // file input stream
  InputStream in;

  // ppm data formats
  static final int PBM_ASCII = 1;
  static final int PGM_ASCII = 2;
  static final int PPM_ASCII = 3;
  static final int PBM_RAW = 4;
  static final int PGM_RAW = 5;
  static final int PPM_RAW = 6;
  int type;

  boolean[] rowsRead;
  int[][] rgbPixels;
  boolean startedRead = false;
  boolean gotSize = false;
  boolean err = false;
  boolean producing = false;
  Vector consumers = new Vector();
  static final ColorModel model = ColorModel.getRGBdefault();

  // image width and height  
  int width = -1, height = -1;

  // maxim value of color (see ppm format documents)
  int maxval;

  /** 
   * Constructor.
   * @param in The stream to read the bytes from.
   */
  public PPMImgProducer(InputStream in) {
    this.in = in;
  }


  //////////////////////////////////////////////////////
  // Methods from ImageProducer.
  //////////////////////////////////////////////////////

  public void addConsumer(ImageConsumer ic) {
    if (ic != null && !isConsumer(ic))
      consumers.addElement(ic);
  }

  public boolean isConsumer(ImageConsumer ic) {
    return consumers.contains(ic);
  }

  public void removeConsumer(ImageConsumer ic) {
    consumers.removeElement( ic );
  }

  public void startProduction(ImageConsumer ic) {
    addConsumer(ic);
    if (!startedRead) {
      startedRead = true;
      new ReadingThread(this);
    }
    if (!producing) {
      producing = true;
      sendImage();
    }
  }

  public void requestTopDownLeftRightResend(ImageConsumer ic) {
    addConsumer(ic);
    waitForSize();
    sendHead(ic);
    for (int row=0; row<height; ++row)
      sendPixelRow(ic, row);
    sendTail( ic );
  }

  ////////////////////////////////////////////////////////////////
  // private methods
  ////////////////////////////////////////////////////////////////
    
  // read header
  private void readHeader(InputStream in) throws IOException {
    char c1 = (char)readByte(in);
    char c2 = (char)readByte(in);

    if (c1 != 'P') throw new IOException( "Not a PBM/PGM/PPM file" );

    switch (c2) {
      case '1':
        type = PBM_ASCII; break;
      case '2':
        type = PGM_ASCII; break;
      case '3':
        type = PPM_ASCII; break;
      case '4':
        type = PBM_RAW;   break;
      case '5':
        type = PGM_RAW;   break;
      case '6':
        type = PPM_RAW;   break;
      default:
        throw new IOException( "Not a standard PBM/PGM/PPM file" );
    }

    width = readInt(in);
    height = readInt(in);
    if (type != PBM_ASCII && type != PBM_RAW)
       maxval = readInt(in);
  }

  // return image width
  private int getWidth() {
    return width;
  }

  // return image heigt
  private int getHeight() {
    return height;
  }

  void readRow(InputStream in, int row, int[] rgbRow) throws IOException {
    int r, g, b;
    int rgb = 0;
    char c;

    for (int col=0; col<width; ++col) {
      switch (type) {
	case PBM_ASCII:
	  c = readChar(in);
  	  if (c == '1')      rgb = 0xff000000;
    	  else if (c == '0') rgb = 0xffffffff;
	  else throw new IOException( "Illegal PBM bit" );
      	  break;

	case PGM_ASCII:
	  g = readInt(in);
	  rgb = makeRgb(g, g, g);
	  break;

	case PPM_ASCII:
	  r = readInt(in);
	  g = readInt(in);
	  b = readInt(in);
	  rgb = makeRgb(r, g, b);
	  break;

	case PBM_RAW:
          if (readBit(in)) rgb = 0xff000000;
          else             rgb = 0xffffffff;
          break;

	case PGM_RAW:
          g = readByte(in);
	  if (maxval != 255) g = fixDepth(g);
          rgb = makeRgb(g, g, g);
 	  break;

	case PPM_RAW:
	  r = readByte(in);
	  g = readByte(in);
	  b = readByte(in);
	  if (maxval != 255) {
	    r = fixDepth(r);
	    g = fixDepth(g);
	    b = fixDepth(b);
          }
	  rgb = makeRgb(r, g, b);
	  break;
      }
      rgbRow[col] = rgb;
    }
  }

  // read the whole image
  void readImage() {
    try {
      readHeader(in);
      width = getWidth();
      height = getHeight();
      if (width == -1 || height == -1)
	err = true;
      else {
	rowsRead = new boolean[height];
	for (int row = 0; row < height; ++row)
	  rowsRead[row] = false;
	gotSize = true;
        synchronized(this) {
          notifyAll();
        }

	rgbPixels = new int[height][width];
	for (int row=0; row<height; ++row) {
          readRow(in, row, rgbPixels[row]);
	  rowsRead[row] = true;
	  synchronized(this) {
            notifyAll();
          }
        }
      }
    } catch (IOException e) {
      err = true;
      width = -1;
      height = -1;
      rowsRead = null;
      rgbPixels = null;
    }
  }

  private void sendImage() {
    int i;

    // Grab the list of consumers, in case it changes while we're sending.
    ImageConsumer[] c = new ImageConsumer[consumers.size()];

    for (i=0; i<c.length; ++i)
      c[i] = (ImageConsumer) consumers.elementAt(i);

    // Try to be as parallel as possible.
    waitForSize();
    for (i=0; i<c.length; ++i)
      sendHead(c[i]);

    for (int row=0; row<height; ++row)
      for (i=0; i<c.length; ++i)
	sendPixelRow( c[i], row );
 
    for (i=0; i<c.length; ++i)
      sendTail(c[i]);

    producing = false;
  }

  private synchronized void waitForSize() {
    while (!err && !gotSize) {
      try {
	wait();
      } catch (InterruptedException ignore) {}
    }
  }

  private synchronized void waitForRow(int row) {
    while (!err && !rowsRead[row]) {
      try {
	wait();
      } catch (InterruptedException ignore) {}
    }
  }

  private void sendHead(ImageConsumer ic) {
    if (err) return;

    ic.setDimensions(width, height);
    ic.setColorModel(model);
    ic.setHints(
	    ImageConsumer.TOPDOWNLEFTRIGHT | ImageConsumer.COMPLETESCANLINES |
	    ImageConsumer.SINGLEPASS | ImageConsumer.SINGLEFRAME );
  }

  private void sendPixelRow(ImageConsumer ic, int row) {
    if (err) return;
    waitForRow(row);
    if (err) return;
    ic.setPixels(0, row, width, 1, model, rgbPixels[row], 0, width);
  }

  private void sendTail(ImageConsumer ic) {
    if (err) ic.imageComplete(ImageConsumer.IMAGEERROR);
    else     ic.imageComplete(ImageConsumer.STATICIMAGEDONE);
  }

  // Utility routine to read a byte.  
  // Instead of returning -1 on EOF, it throws an exception.
  private static int readByte(InputStream in) throws IOException {
    int b = in.read();
    if (b == -1) throw new EOFException();
    return b;
  }

  private int bitshift = -1;
  private int bits;

  // Utility routine to read a bit, packed eight to a byte, big-endian.
  private boolean readBit(InputStream in) throws IOException {
    if (bitshift == -1) {
      bits = readByte(in); 
      bitshift = 7;
    }
    boolean bit = ((( bits>>bitshift) & 1) != 0);
    --bitshift;
    return bit;
  }

  // Utility routine to read a character, ignoring comments.
  private static char readChar(InputStream in) throws IOException {
    char c = (char) readByte(in);
    if (c == '#') {
      do {
	c = (char) readByte(in);
      } while (c != '\n' && c != '\r');
    }
    return c;
  }

  // Utility routine to read the first non-whitespace character.
  private static char readNonwhiteChar(InputStream in) throws IOException {
    char c;

    do {
      c = readChar(in);
    } while (c == ' ' || c == '\t' || c == '\n' || c == '\r');

    return c;
  }

  // Utility routine to read an ASCII integer, ignoring comments.
  private static int readInt(InputStream in) throws IOException {
    char c = readNonwhiteChar(in);
    if (c < '0' || c > '9')
      throw new IOException( "Junk in file where integer should be." );

    int i = 0;
    do {
      i = i * 10 + c - '0';
      c = readChar( in );
    } while (c >= '0' && c <= '9');

    return i;
  }

  // Utility routine to rescale a pixel value from a non-eight-bit maxval.
  private int fixDepth(int p) {
    return (p * 255 + maxval/2 ) / maxval;
  }

  // Utility routine make an RGBdefault pixel from three color values.
  private static int makeRgb(int r, int g, int b) {
    return 0xff000000 | ( r << 16 ) | ( g << 8 ) | b;
  }
}


class ReadingThread extends Thread {

  private PPMImgProducer parent;

  public ReadingThread(PPMImgProducer parent) {
    this.parent = parent;
    start();
  }

  // Methods from Runnable.
  public void run() {
    parent.readImage();
  }
}
