package sv.viewer;
               
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import com.sun.java.swing.*;
import sv.engine.*;
import sv.gui.*;
import sv.data.*;
import sv.config.*;
import sv.util.*;

/** 
 * main frame to visualize the data
 */

// An abstract class is a class that is partially implemented 
// and whose purpose is solely as a design convenience. Abstract class
// are made up of one or more abstract methods, which are methods
// that are declared but left bodiless.

public abstract class GFrame extends Frame {

  public static final int RGBCOLORMODEL = 1;
  public static final int HSBCOLORMODEL = 2;

  /**
   * id number of the visualization window
   */
  protected long identification;

  /** 
   * to display the information of the graph and error message
   */
  protected StatusBar status;

  /**
   * data set to this viewer
   */
  protected DataSetBrief dataSet; 

  /**
   * Time value array for all dataset attached this frame
   */
  protected float[] timeValues;

  /**
   * indicate the current time
   */
  protected int currentTime = 0;

  /**
   * configuration
   */
  protected Configs config;

  /**
   * common menu items
   */
  protected MenuItem animateMI, nextMI, prevMI, stopMI;

  /**
   * animation thread
   */
  protected Animator animator;
 
  /**
   * animation speed
   */
  protected int animationSpeed = 200;

  /**
   * filter menu 
   */
  protected Menu filterMenu;

  /**
   * overlay and discard memu item
   */
  protected CheckboxMenuItem overlayMI, discardMI;

  // for color map
  public int colorModel  = 1;
  public NewtonMethod rspline, gspline, bspline, 
                      hspline, sspline, lspline;  


  /**
   * default constructor
   */
  public GFrame() {
  }

  /**
   * constructor(to create peer frame for split command)
   * @param brief data set to be visualized
   */
  public GFrame(DataSetBrief brief, Configs cfg) {
    super(brief.getDataName());
    dataSet = brief;
    config = cfg;

    boolean b = false;
    if (config.getColormap() != null) {
      b = readColorMap(config.getColormap());

      if (b) {
        float value;

        for (int i=0; i<101; i++) {
           value = rspline.computePoint((float)i/100f); 
           if (value > 1.0f) value = 1.0f;
           else if (value < 0.0f) value = 0.0f;
           rcolors[i] = value;

           value = gspline.computePoint((float)i/100f);    
           if (value > 1.0f) value = 1.0f;
           else if (value < 0.0f) value = 0.0f;
           gcolors[i] = value;

           value = bspline.computePoint((float)i/100f);    
           if (value > 1.0f) value = 1.0f;
           else if (value < 0.0f) value = 0.0f;
           bcolors[i] = value;
        }
      }
    }

    if (config.getColormap() == null || b == false) {
      int nControls = 5;
      float[] xcontrols = { 0.0f, 0.25f, 0.5f, 0.75f, 1.0f };
      float[] ycontrols = { 0.0f, 0.5f, 1.0f, 0.5f, 0.0f };
      rspline = new NewtonMethod(xcontrols, ycontrols);
      hspline = new NewtonMethod(xcontrols, ycontrols);

      for (int i=0; i<5; i++)
        ycontrols[i] = 0.5f;
      gspline = new NewtonMethod(xcontrols, ycontrols);
      sspline = new NewtonMethod(xcontrols, ycontrols);

      ycontrols[0] = 1.0f;
      ycontrols[1] = 0.5f; 
      ycontrols[2] = 0.0f;
      ycontrols[3] = 0.5f;
      ycontrols[4] = 1.0f;
      bspline = new NewtonMethod(xcontrols, ycontrols);
      lspline = new NewtonMethod(xcontrols, ycontrols);
    }

    updateTimeValues();
  }


  /**
   * update time values of this viewer window
   */
  public void updateTimeValues() {
    DataSet[] data = dataSet.getDataSet();
    Set s = new TreeSet();

    for (int i=0; i<data.length; i++) {
      int sz = data[i].getNoOfTimes();
      for (int j=0; j<sz; j++) 
        s.add(new Float(data[i].getTimeValue(j).getAbsTime()));
    }
    Object[] tmp = s.toArray();
    timeValues = new float[tmp.length];
    for (int i=0; i<tmp.length; i++) 
      timeValues[i] = ((Float)tmp[i]).floatValue(); 
  }   

  /**
   * return id of this visualization window
   */
  public long getID() {
    return identification;
  }

  /**
   * set id for this visualization window
   * @param id id for this visualization window
   */
  public void setID(long id) {
    identification = id;
  }

  /**
   * set title of this frame
   * @param title title of this frame
   */
  public void setFrameTitle(String title) {
     setTitle(title);
  }

  /**
   * return data set
   */
  public DataSetBrief getDataSet() {
    return dataSet;
  }

  /**
   * return current time
   */
  public float getCurrent() {
    return timeValues[currentTime];
  }

  /**
   * return current index
   */
  public int getCurrentIndex() {
    return currentTime;
  }

  /**
   * change color map info
   */
  public abstract void changeColorMap(String[] args);

  /**
   * repaint
   */
  public abstract void notifyRepaint();

  /**
   * do action event
   */
  public abstract void doActionEventWork(byte[] datum);

  /**
   * update data set
   */
  public abstract void updateDataSet();

  /**
   * animate 
   */
  public abstract void doAnimate();

  /**
   * next
   */
  public abstract void doNext(boolean b);

  /**
   * previous 
   */
  public abstract void doPrevious();

  /**
   * stop 
   */
  public abstract void doStop(int pos);

  /**
   * close this visualization window
   */
  public abstract void closeFrame();

  /**
   * write misc information into the file
   */
  public abstract void writeAux(DataOutputStream out) throws IOException;

  /**
   * send the action event message to the collaborator
   * @param comm command with string type
   */
  public void catchActionEvent(String comm) {
    if (!dataSet.isPrivate()) {

      byte[] datum = comm.getBytes();
     
      try {
        config.getEngine().sender.broadcastCommand(datum);
      } catch (Exception e) {
        System.err.println("Error: connection to server");
      }

      if (config.getEngine().sender.getIsMaster()) 
        doActionEventWork(comm.getBytes());
      else
        config.getEngine().getGUI().message("(Slave)Event action to shared data was disabled.", true);
    } else
      doActionEventWork(comm.getBytes());
  }

  /**
   * set color map
   */
  protected void setColorMap(StringTokenizer st) {
     colorModel = Integer.valueOf(st.nextToken()).intValue();
     for (int i=0; i<101; i++) {
        rcolors[i] = Float.valueOf(st.nextToken()).floatValue();
        gcolors[i] = Float.valueOf(st.nextToken()).floatValue();
        bcolors[i] = Float.valueOf(st.nextToken()).floatValue();
        hcolors[i] = Float.valueOf(st.nextToken()).floatValue();
        scolors[i] = Float.valueOf(st.nextToken()).floatValue();
        lcolors[i] = Float.valueOf(st.nextToken()).floatValue();
     }
  }

  /**
   * read the color map file 
   */
  protected boolean readColorMap(String fname) {

    try {
       BufferedReader reader = new BufferedReader(new FileReader(fname));

       int nControls;
       float[] xcontrols, ycontrols;
       String input, type;
       StringTokenizer st;

       for (int tt=0; tt<6; tt++) {
          input = reader.readLine();
          st = new StringTokenizer(input);
          type = st.nextToken().toLowerCase();

          if (type.indexOf("r") >= 0) {
             nControls = Integer.valueOf(st.nextToken()).intValue();

             xcontrols = new float[nControls];
             ycontrols = new float[nControls];

             for (int i=0; i<nControls; i++) {
               input = reader.readLine();
		st = new StringTokenizer(input);

               xcontrols[i] = Float.valueOf(st.nextToken()).floatValue();
               ycontrols[i] = Float.valueOf(st.nextToken()).floatValue();
             }
  
             rspline = new NewtonMethod(xcontrols, ycontrols);
             
          } else if (type.indexOf("g") >= 0) {
             nControls = Integer.valueOf(st.nextToken()).intValue();

             xcontrols = new float[nControls];
             ycontrols = new float[nControls];

             for (int i=0; i<nControls; i++) {
               input = reader.readLine();
		st = new StringTokenizer(input);

               xcontrols[i] = Float.valueOf(st.nextToken()).floatValue();
               ycontrols[i] = Float.valueOf(st.nextToken()).floatValue();
             }
  
             gspline = new NewtonMethod(xcontrols, ycontrols);

          } else if (type.indexOf("b") >= 0) {
             nControls = Integer.valueOf(st.nextToken()).intValue();

             xcontrols = new float[nControls];
             ycontrols = new float[nControls];

             for (int i=0; i<nControls; i++) {
               input = reader.readLine();
		st = new StringTokenizer(input);

               xcontrols[i] = Float.valueOf(st.nextToken()).floatValue();
               ycontrols[i] = Float.valueOf(st.nextToken()).floatValue();
             }
  
             bspline = new NewtonMethod(xcontrols, ycontrols);

          } else if (type.indexOf("h") >= 0) {
             nControls = Integer.valueOf(st.nextToken()).intValue();

             xcontrols = new float[nControls];
             ycontrols = new float[nControls];

             for (int i=0; i<nControls; i++) {
               input = reader.readLine();
		st = new StringTokenizer(input);

               xcontrols[i] = Float.valueOf(st.nextToken()).floatValue();
               ycontrols[i] = Float.valueOf(st.nextToken()).floatValue();
             }
  
             hspline = new NewtonMethod(xcontrols, ycontrols);

          } else if (type.indexOf("s") >= 0) {
             nControls = Integer.valueOf(st.nextToken()).intValue();

             xcontrols = new float[nControls];
             ycontrols = new float[nControls];

             for (int i=0; i<nControls; i++) {
               input = reader.readLine();
		st = new StringTokenizer(input);

               xcontrols[i] = Float.valueOf(st.nextToken()).floatValue();
               ycontrols[i] = Float.valueOf(st.nextToken()).floatValue();
             }
  
             sspline = new NewtonMethod(xcontrols, ycontrols);

          } else if (type.indexOf("l") >= 0) {
             nControls = Integer.valueOf(st.nextToken()).intValue();

             xcontrols = new float[nControls];
             ycontrols = new float[nControls];

             for (int i=0; i<nControls; i++) {
               input = reader.readLine();
		st = new StringTokenizer(input);

               xcontrols[i] = Float.valueOf(st.nextToken()).floatValue();
               ycontrols[i] = Float.valueOf(st.nextToken()).floatValue();
             }
  
             lspline = new NewtonMethod(xcontrols, ycontrols);

          } else {
            System.out.println("Can't read colormap file...");
            reader.close();
            return false;
          }
       }
       reader.close();
       return true;

    } catch (Exception e) {
       System.out.println("Error while reading the colormap file.");
       return false;
    }
  }

  // create the filter menu for the overlay and discard menu item
  protected void subFilterMenu() {
    overlayMI = new CheckboxMenuItem("Overlay Mode");
    overlayMI.setState(config.isOverlayMode());
    overlayMI.addItemListener(new ItemListener() {
      public void itemStateChanged(ItemEvent e) {
        config.setOverlayMode(overlayMI.getState());
      }
    });
    filterMenu.add(overlayMI);

    discardMI = new CheckboxMenuItem("Discard Mode");
    discardMI.setState(config.isDiscardMode());
    discardMI.addItemListener(new ItemListener() {
      public void itemStateChanged(ItemEvent e) {
        config.setDiscardMode(discardMI.getState());
      }
    });
    filterMenu.add(discardMI);
  }

  /**
   * do filter action
   * @param info filter menu information
   */
  public void doFilterAction(FilterMenuInfo info) {

    if (!config.getEngine().sender.getIsMaster()) {
      config.getEngine().getGUI().message("(Slave)Event action to shared data was disabled.", true);
      return;
    }

    Filter filter = null;
    String dir = "";
    String arg = "";

    try {
      if (info.directory != null)
         dir = info.directory;
      filter=(Filter)Class.forName(dir+"."+info.className).newInstance();
    } catch (Exception exception) {
      System.out.println("Can't find " + info.className + "...");
      return;
    }

    if (filter.getStyle() == Filter.NORMAL) {
      if (filter.getNoOfArgs() != 0) 
        arg = JOptionPane.showInputDialog(this, "Enter the arguments:");

    } else if (filter.getStyle() == Filter.MULTIPLE) {
      config.getEngine().vmanager.startMerge(filter, dataSet, 
                              config.isOverlayMode(), config.isDiscardMode());
      config.getEngine().getGUI().iconDisplay.setMergeOption(true);
      status.display("Select the second data set from icon panel.", false);
      return;
    }

    if (config.getEngine().sender.isCollabarable() && !dataSet.isPrivate()) {
System.out.println("Collaboration filter");
 
      boolean bycodes = false;
      byte[] result = null;
      String fname = null;
      int size;

      try {
        Properties dprop = System.getProperties();
        String file_seperator = dprop.getProperty("file.separator");
        String java_path = dprop.getProperty("java.class.path");
        String path_separator = dprop.getProperty("path.separator");
        StringTokenizer st = new StringTokenizer(java_path, path_separator);
        String classname = filter.getClass().getName();
        StringTokenizer st2 = new StringTokenizer(classname, ".");

        String cname = "";
        size = st2.countTokens(); 
        for (int i=0; i<size-1; i++) 
          cname += st2.nextToken() + file_seperator;

        fname = st2.nextToken();
        cname += (fname + ".class");

        String path;
        size = st.countTokens();
        for (int i=0; i<size; i++) {
          path = st.nextToken();

          if (path.indexOf(".jar") > 0) continue;

          File target = new File(path, cname);
          if (target == null) continue;

          if (!target.exists()) continue;
          else {
            bycodes = true;
            DataInputStream in = new DataInputStream(new FileInputStream(target));
            int sz = in.available();
            result = new byte[sz];
    	    in.readFully(result, 0, sz); 
            break;
          }
	}
      } catch (Exception E) {
         bycodes = false; 
      }

      if (!bycodes) {
        status.display("Can't find the filter codes.", true);
        return;
      }
      
      try {
        config.getEngine().sender.serverFilterAction(
                  dataSet.getID(), fname, result, arg, -1, 
                  String.valueOf(config.isOverlayMode()),
                  String.valueOf(config.isDiscardMode()));
      } catch (Exception e) {
        status.display("Filter broadcasting error.", true);
        return;
      }

    } else {
System.out.println("Non-collaboration filter");

      DataSet[] ds = dataSet.getDataSet();
      int noOfInputs = filter.getNoOfArgs();
      DataSetBrief brief = null;

      for (int elt=0; elt<ds.length; elt++) {
        DataSet data = ds[elt];
        filter.clear();

        if (noOfInputs == 1) {
           float val = 0.0f;
           try {
              val = Float.valueOf(arg).floatValue();
           } catch (Exception ee) {
              status.display("Input should be number...", false);
              return;
           }

           filter.assignDataSet(data);
           filter.assignArgs(val);

        } else if (noOfInputs == 2) {
           float start = 0, end = 0;
           int where = arg.indexOf("-");
           if (where < 0) {
              status.display("Usage: start - end", true);
              return;
           } else {
              try {
                start = 
                 Float.valueOf(arg.substring(0, where).trim()).floatValue();
                end = 
                 Float.valueOf(arg.substring(where+1).trim()).floatValue();
              } catch (Exception ee) {
                  status.display("Usage: start - end", false);
                  return;
              }

              if (end < start) {
                  status.display("Invalid range ...", true);
                  return;
              }
           }
           filter.assignDataSet(data);
           filter.assignArgs(start, end);

        } else         // if it doesn't need an argument
           filter.assignDataSet(data);

        DataSet newdata;
        try { 
          newdata = (DataSet)filter.performFilter();
	} catch (Exception ee) {
          status.display("Filter can't perform. Check filter codes.", true);
          return;
        }

        if (newdata == null) {
          status.display("Filter can't perform. Check filter codes.", true);
          return;
        }

        if (config.isOverlayMode()) {
	   dataSet.addDataSet(newdata);

        } else {
          if (brief == null)
            brief = new DataSetBrief(newdata.getDataName(), 
                                     newdata.getNoOfTimes(), 
                                     filter.getOutDimension(),
                                     newdata.getDataSize());

          brief.addDataSet(newdata);
        }
      }

      if (brief != null) {
        brief.setID(System.currentTimeMillis());
        brief.updateInfo();
        brief.setPrivate(true);
        config.getEngine().receivedBriefInfo(brief); 
        long fid = config.getEngine().vmanager.register(brief, -1, null);
        brief.setViewerID(fid); 
      }

      if (config.isDiscardMode()) 
        config.getEngine().vmanager.deleteDataSet(dataSet);

      notifyRepaint();
    }
  }

  // close action
  protected ActionListener closeListener = new ActionListener() {
    public void actionPerformed(ActionEvent e) {
      if (!dataSet.isPrivate()) {
        String comm = String.valueOf(SV2Action.CLOSE) + ";" +
                                       String.valueOf(dataSet.getID());

        byte[] datum = comm.getBytes();
       
        try {
          config.getEngine().sender.serverCommand(datum);
          config.getEngine().sender.broadcastCommand(datum);
        } catch (Exception E) {
          System.err.println("Error: connection to server");
        }

        if (config.getEngine().sender.getIsMaster()) 
          config.getEngine().vmanager.deleteDataSet(dataSet);
        else
          config.getEngine().getGUI().message("(Slave)Event action to shared data was disabled.", true);
      } else
        config.getEngine().vmanager.deleteDataSet(dataSet);
    }
  };

  // animate action
  protected ActionListener animateListener = new ActionListener() {
    public void actionPerformed(ActionEvent e) {
      if (!dataSet.isPrivate()) {
        String comm = String.valueOf(SV2Action.ANIMATE) + ";" +
                                       String.valueOf(dataSet.getID());

        byte[] datum = comm.getBytes();
       
        try {
          config.getEngine().sender.broadcastCommand(datum);
        } catch (Exception E) {
          System.err.println("Error: connection to server");
        }

        if (config.getEngine().sender.getIsMaster()) 
          config.getEngine().vmanager.animateAction(identification);
        else
          config.getEngine().getGUI().message("(Slave)Event action to shared data was disabled.", true);
      } else
        config.getEngine().vmanager.animateAction(identification);
    }
  };

  // next action
  protected ActionListener nextListener = new ActionListener() {
    public void actionPerformed(ActionEvent e) {
      if (!dataSet.isPrivate()) {
        String comm = String.valueOf(SV2Action.NEXT) + ";" +
                                       String.valueOf(dataSet.getID());

        byte[] datum = comm.getBytes();
       
        try {
          config.getEngine().sender.broadcastCommand(datum);
        } catch (Exception E) {
          System.err.println("Error: connection to server");
        }

        if (config.getEngine().sender.getIsMaster()) 
          config.getEngine().vmanager.nextAction(identification);
        else
          config.getEngine().getGUI().message("(Slave)Event action to shared data was disabled.", true);
      } else
        config.getEngine().vmanager.nextAction(identification);
    }
  };

  // previous action
  protected ActionListener previousListener = new ActionListener() {
    public void actionPerformed(ActionEvent e) {
      if (!dataSet.isPrivate()) {
        String comm = String.valueOf(SV2Action.PREVIOUS) + ";" +
                                       String.valueOf(dataSet.getID());

        byte[] datum = comm.getBytes();
       
        try {
          config.getEngine().sender.broadcastCommand(datum);
        } catch (Exception E) {
          System.err.println("Error: connection to server");
        }

        if (config.getEngine().sender.getIsMaster()) 
          config.getEngine().vmanager.previousAction(identification);
        else
          config.getEngine().getGUI().message("(Slave)Event action to shared data was disabled.", true);
      } else
        config.getEngine().vmanager.previousAction(identification);
    }
  };

  // stop action
  protected ActionListener stopListener = new ActionListener() {
    public void actionPerformed(ActionEvent e) {
      int current = currentTime; 
      if (!dataSet.isPrivate()) {
        String comm = String.valueOf(SV2Action.STOP) + ";" +
                 String.valueOf(dataSet.getID()) + ";" + 
                 String.valueOf(current);

        byte[] datum = comm.getBytes();
       
        try {
          config.getEngine().sender.broadcastCommand(datum);
        } catch (Exception E) {
          System.err.println("Error: connection to server");
        }

        if (config.getEngine().sender.getIsMaster()) 
          config.getEngine().vmanager.stopAction(identification, current);
        else
          config.getEngine().getGUI().message("(Slave)Event action to shared data was disabled.", true);
      } else
        config.getEngine().vmanager.stopAction(identification, current);
    }
  };

  // save as action
  protected ActionListener saveasListener = new ActionListener() {
    public void actionPerformed(ActionEvent event) {
      FileDialog openFileDialog = new FileDialog(GFrame.this,
                                 "Save DataSet", FileDialog.SAVE);
      openFileDialog.show();
      String input = openFileDialog.getFile();
      if (input == null) return;

      boolean b = input.indexOf(".") >=0 ? true : false; 
      if (!b)  input += ".sv2";
      input = openFileDialog.getDirectory() + input;

      DataOutputStream os = null;
      try {
        os = new DataOutputStream(
               new BufferedOutputStream(new FileOutputStream(input)));
        dataSet.writeToFile(os);
	writeAux(os);
        os.close();

      } catch (Exception e) {
e.printStackTrace();
        System.out.println("File writing error...");
        return;
      }
    }
  };

public float rcolors[] = { 
0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.02f, 0.03f, 
0.05f, 0.07f, 0.10f, 0.12f, 0.15f, 0.18f, 0.21f, 0.24f, 0.27f, 0.30f, 
0.33f, 0.37f, 0.40f, 0.43f, 0.47f, 0.50f, 0.53f, 0.57f, 0.60f, 0.63f, 
0.66f, 0.69f, 0.72f, 0.75f, 0.78f, 0.80f, 0.83f, 0.85f, 0.87f, 0.89f, 
0.91f, 0.93f, 0.94f, 0.95f, 0.97f, 0.98f, 0.99f, 0.99f, 1.00f, 1.00f,
1.00f, 1.00f, 1.00f, 0.99f, 0.99f, 0.98f, 0.97f, 0.95f, 0.94f, 0.93f, 
0.91f, 0.89f, 0.87f, 0.85f, 0.83f, 0.80f, 0.78f, 0.75f, 0.72f, 0.70f, 
0.66f, 0.63f, 0.60f, 0.57f, 0.53f, 0.50f, 0.47f, 0.43f, 0.40f, 0.37f, 
0.33f, 0.30f, 0.27f, 0.24f, 0.21f, 0.18f, 0.15f, 0.12f, 0.10f, 0.07f, 
0.05f, 0.03f, 0.02f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 
0.0f};

public float gcolors[] = { 
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f };

public float bcolors[] = { 
1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 0.98f, 0.97f, 
0.95f, 0.93f, 0.90f, 0.88f, 0.85f, 0.82f, 0.79f, 0.76f, 0.73f, 0.70f, 
0.67f, 0.63f, 0.60f, 0.57f, 0.53f, 0.50f, 0.47f, 0.43f, 0.40f, 0.37f, 
0.34f, 0.31f, 0.28f, 0.25f, 0.22f, 0.20f, 0.17f, 0.15f, 0.13f, 0.11f, 
0.09f, 0.07f, 0.06f, 0.05f, 0.03f, 0.02f, 0.01f, 0.01f, 0.00f, 0.00f,
0.00f, 0.00f, 0.00f, 0.01f, 0.01f, 0.02f, 0.03f, 0.05f, 0.06f, 0.07f, 
0.09f, 0.11f, 0.13f, 0.15f, 0.17f, 0.20f, 0.22f, 0.25f, 0.28f, 0.31f, 
0.34f, 0.37f, 0.40f, 0.43f, 0.47f, 0.50f, 0.53f, 0.57f, 0.60f, 0.63f, 
0.67f, 0.70f, 0.73f, 0.76f, 0.79f, 0.82f, 0.85f, 0.88f, 0.90f, 0.93f, 
0.95f, 0.97f, 0.98f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f,
1.00f }; 


public float hcolors[] = { 
0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.02f, 0.03f, 
0.05f, 0.07f, 0.10f, 0.12f, 0.15f, 0.18f, 0.21f, 0.24f, 0.27f, 0.30f, 
0.33f, 0.37f, 0.40f, 0.43f, 0.47f, 0.50f, 0.53f, 0.57f, 0.60f, 0.63f, 
0.66f, 0.69f, 0.72f, 0.75f, 0.78f, 0.80f, 0.83f, 0.85f, 0.87f, 0.89f, 
0.91f, 0.93f, 0.94f, 0.95f, 0.97f, 0.98f, 0.99f, 0.99f, 1.00f, 1.00f,
1.00f, 1.00f, 1.00f, 0.99f, 0.99f, 0.98f, 0.97f, 0.95f, 0.94f, 0.93f, 
0.91f, 0.89f, 0.87f, 0.85f, 0.83f, 0.80f, 0.78f, 0.75f, 0.72f, 0.70f, 
0.66f, 0.63f, 0.60f, 0.57f, 0.53f, 0.50f, 0.47f, 0.43f, 0.40f, 0.37f, 
0.33f, 0.30f, 0.27f, 0.24f, 0.21f, 0.18f, 0.15f, 0.12f, 0.10f, 0.07f, 
0.05f, 0.03f, 0.02f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 
0.0f};

public float scolors[] = { 
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f };

public float lcolors[] = { 
1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 0.98f, 0.97f, 
0.95f, 0.93f, 0.90f, 0.88f, 0.85f, 0.82f, 0.79f, 0.76f, 0.73f, 0.70f, 
0.67f, 0.63f, 0.60f, 0.57f, 0.53f, 0.50f, 0.47f, 0.43f, 0.40f, 0.37f, 
0.34f, 0.31f, 0.28f, 0.25f, 0.22f, 0.20f, 0.17f, 0.15f, 0.13f, 0.11f, 
0.09f, 0.07f, 0.06f, 0.05f, 0.03f, 0.02f, 0.01f, 0.01f, 0.00f, 0.00f,
0.00f, 0.00f, 0.00f, 0.01f, 0.01f, 0.02f, 0.03f, 0.05f, 0.06f, 0.07f, 
0.09f, 0.11f, 0.13f, 0.15f, 0.17f, 0.20f, 0.22f, 0.25f, 0.28f, 0.31f, 
0.34f, 0.37f, 0.40f, 0.43f, 0.47f, 0.50f, 0.53f, 0.57f, 0.60f, 0.63f, 
0.67f, 0.70f, 0.73f, 0.76f, 0.79f, 0.82f, 0.85f, 0.88f, 0.90f, 0.93f, 
0.95f, 0.97f, 0.98f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f,
1.00f }; 

} // end of class
