package sv.viewer;

import java.util.*;
import java.awt.*;
import java.io.*;
import sv.data.*;
import sv.engine.*;
import sv.config.*;
import sv.viewer.builtin.*;
import sv.util.*;

public class ViewManager {
  
  // viewer vector
  private Vector frameVector;

  // friend class which manages clients' work
  static SV2Engine engine;

  // configurations
  private ClientConfig config;

  // viewer option
  private int viewer_opt;    

  // Filter menu handler
  private FilterMenuHandler filterMenu;

  // clipboard data set
  private DataSetBrief clipboard = null;

  // merge target data set
  private MergeHandle mergeHandle = null;

  /**
   * constructor
   */
  public ViewManager(SV2Engine eng, int opt, String dir) {
    engine = eng;
    viewer_opt = opt;
    frameVector = new Vector();
    config = new ClientConfig(eng, dir);
    filterMenu = new FilterMenuHandler(config.getMenuFile());
  }

  /**
   * return filter menu
   */
  public FilterMenuHandler getFilterMenu() {
    return filterMenu;
  } 

  /**
   * return if the matched viewer exists for the given viewre id
   */
  public GFrame isExist(long id) {
    for (Enumeration e=frameVector.elements(); e.hasMoreElements();) {
      GFrame gframe = (GFrame)e.nextElement();
      if (gframe.getID() == id) 
        return gframe;
    }
    return null;
  }

  /**
   * set clipboard
   * @param brief intended data set for the clipboard used for cut & copy
   */
  public void setClipboard(DataSetBrief brief) {
    clipboard = new DataSetBrief(brief.getDataName(), brief.getNoOfTime(),
                                 brief.getRank(), brief.getDataSize());

    clipboard.setID(brief.getID());
    clipboard.setViewerID(brief.getViewerID());
    clipboard.setPrivate(brief.isPrivate());
    DataSet[] ds = brief.getDataSet();
    if (ds != null) 
      for (int i=0; i<ds.length; i++)
        clipboard.addDataSet(ds[i].getClone());
  }

  /**
   * get clipboard
   */
  public DataSetBrief getClipboard() {
    return clipboard;
  }


  /**
   * start merge action 
   */
  public void startMerge(Filter filter, DataSetBrief brief,
                              boolean overlay, boolean discard) {
    mergeHandle = new MergeHandle(filter, brief, overlay, discard);
    
  }

  /**
   * real merge action
   */
  public void mergeAction(DataSetBrief brief) {
    if (mergeHandle == null) {
      engine.getGUI().message("Merging failed...", true);
      return;
    }

    if (brief.isPrivate()) 
      if (!mergeHandle.target.isPrivate()) {
        engine.getGUI().message("Merging shared data and non-shared data is not allowed.", true);
        return;
      }

    if (mergeHandle.target.isPrivate()) 
      if (!brief.isPrivate()) {
        engine.getGUI().message("Merging shared data and non-shared data is not allowed.", true);
        return;
      }

    Filter filter = mergeHandle.filter;

    if (engine.sender.isCollabarable() && 
             !brief.isPrivate() && !mergeHandle.target.isPrivate()) {
System.out.println("Collaboration merging filter");
 
      boolean bycodes = false;
      byte[] result = null;
      String fname = null;
      int size;

      try {
        Properties dprop = System.getProperties();
        String file_seperator = dprop.getProperty("file.separator");
        String java_path = dprop.getProperty("java.class.path");
        String path_separator = dprop.getProperty("path.separator");
        StringTokenizer st = new StringTokenizer(java_path, path_separator);
        String classname = filter.getClass().getName();
        StringTokenizer st2 = new StringTokenizer(classname, ".");

        String cname = "";
        size = st2.countTokens(); 
        for (int i=0; i<size-1; i++) 
          cname += st2.nextToken() + file_seperator;

        fname = st2.nextToken();
        cname += (fname + ".class");

        String path;
        size = st.countTokens();
        for (int i=0; i<size; i++) {
          path = st.nextToken();

          if (path.indexOf(".jar") > 0) continue;

          File target = new File(path, cname);
          if (target == null) continue;

          if (!target.exists()) continue;
          else {
            bycodes = true;
            DataInputStream in = new DataInputStream(new FileInputStream(target));
            int sz = in.available();
            result = new byte[sz];
    	    in.readFully(result, 0, sz); 
            break;
          }
	}
      } catch (Exception E) {
         bycodes = false; 
      }

      if (!bycodes) {
        engine.getGUI().message("Can't find the filter codes.", true);
        return;
      }
      
      try {
        engine.sender.serverFilterAction(
                mergeHandle.target.getID(), fname, result, "", brief.getID(), 
                String.valueOf(mergeHandle.overlay),
                String.valueOf(mergeHandle.discard));
      } catch (Exception e) {
        engine.getGUI().message("Filter broadcasting error.", true);
        return;
      }

    } else {
System.out.println("Non-collaboration merging filter");
      DataSet[] dss1 = mergeHandle.target.getDataSet();
      DataSet[] dss2 = brief.getDataSet();

      if (dss1.length!=1 || dss2.length!=1) {
        System.out.println("Multi-data merging does not support yet.");
        return;
      }
     
      int noOfInputs = filter.getNoOfArgs();
      DataSetBrief res_data = null;

      filter.clear();

      if (noOfInputs >= 1) {
        System.out.println("Merging with arguments does not support yet.");
      } else {        // if it doesn't need an argument
        filter.assignDataSet(dss1[0]);
        filter.assignDataSet(dss2[0]);
      }

      DataSet newdata;
      try { 
        newdata = (DataSet)filter.performFilter();
      } catch (Exception ee) {
        engine.getGUI().message("Filter can't perform. Check filter codes.", true);
        return;
      }

      if (newdata == null) {
        engine.getGUI().message("Filter can't perform. Check filter codes.", true);
        return;
      }

      if (mergeHandle.overlay) {
        mergeHandle.target.addDataSet(newdata);

        if (mergeHandle.target.getViewerID() > 0) {
          long vid = mergeHandle.target.getViewerID();
          for (int i=0; i<frameVector.size(); i++) {
            GFrame gframe = (GFrame)frameVector.elementAt(i);
            if (gframe.getID() == vid) {
              gframe.notifyRepaint();
              break;
            }
          }
        }


      } else {
        if (res_data == null)
          res_data = new DataSetBrief(newdata.getDataName(), 
                                   newdata.getNoOfTimes(), 
                                   filter.getOutDimension(),
                                   newdata.getDataSize());

        res_data.addDataSet(newdata);
      }

      if (res_data != null) {
        res_data.setID(System.currentTimeMillis());
        res_data.updateInfo();
        res_data.setPrivate(true);
        engine.receivedBriefInfo(res_data); 
        long fid = engine.vmanager.register(res_data, -1, null);
        res_data.setViewerID(fid); 
      }

      if (mergeHandle.discard) {
        System.out.println("Discarding original data for merging does not support yet.");
      }
    }
  }

  /**
   * filter processing function in the server
   * @param id1 first data set identification
   * @param id2 second data set identification, used only for merging
   * @param filter_name filter name
   * @param argument filter argument
   * @param filter_code filter source code
   * @param overlay overlay mode
   * @param discard discard mode
   * @param did new data set id
   * @param vid new data set viewer id
   */
  public void generalFilterWork(long id1, long id2, String filter_name, 
                                String argument, byte[] filter_code, 
                                boolean overlay, boolean discard,
                                long did, long vid) {

    System.out.println("Client filter action: " + filter_name);
    DataSetBrief brief = null;

    if (filter_code.length < 1) {
      System.out.println("Client filter action error(null).");
      return;
    } 
   
    try {
      FilterClassLoader loader = new FilterClassLoader(filter_code);
      Class rfilter = loader.loadClass(filter_name, true);
      if (rfilter == null) {
        System.out.println("Client filter action error(Format).");
        return;
      }

      Filter filter = (Filter)rfilter.newInstance();

      if (filter.getStyle() == Filter.NORMAL) {
        DataSetBrief dataSet = 
                       engine.getGUI().iconDisplay.getDataSetBrief(id1);
        DataSet[] ds = dataSet.getDataSet();
        int noOfInputs = filter.getNoOfArgs();

        for (int elt=0; elt<ds.length; elt++) {
          DataSet data = ds[elt];
          filter.clear();

          if (noOfInputs == 1) {
             float val = 0.0f;
             try {
                val = Float.valueOf(argument).floatValue();
             } catch (Exception ee) {
                System.out.println("Client filter action error(Input).");
                return;
             }

             filter.assignDataSet(data);
             filter.assignArgs(val);

          } else if (noOfInputs == 2) {
             float start = 0, end = 0;
             int where = argument.indexOf("-");
             if (where < 0) {
                System.out.println("Client filter action error(Usage).");
                return;
             } else {
                try {
                  start = 
                   Float.valueOf(argument.substring(0, where).trim()).floatValue();
                  end = 
                   Float.valueOf(argument.substring(where+1).trim()).floatValue();
                } catch (Exception ee) {
                    System.out.println("Client filter action error(Usage).");
                    return;
                }

                if (end < start) {
                  System.out.println("Client filter action error(Range).");
                  return;
                }
             }
             filter.assignDataSet(data);
             filter.assignArgs(start, end);

          } else         // if it doesn't need an argument
             filter.assignDataSet(data);

          DataSet newdata;
          try { 
            newdata = (DataSet)filter.performFilter();
  	  } catch (Exception ee) {
            System.out.println("Client filter action error(Code).");
            return;
          }

          if (newdata == null) {
            System.out.println("Client filter action error(null result).");
            return;
          }

          if (overlay) {
	    dataSet.addDataSet(newdata);

            if (dataSet.getViewerID() > 0) {
              long vid2 = dataSet.getViewerID();
              for (int i=0; i<frameVector.size(); i++) {
                GFrame gframe = (GFrame)frameVector.elementAt(i);
                if (gframe.getID() == vid2) {
                  gframe.notifyRepaint();
                  break;
                }
              }
            }

          } else {
            if (brief == null)
              brief = new DataSetBrief(newdata.getDataName(), 
                                     newdata.getNoOfTimes(), 
                                     filter.getOutDimension(),
                                     newdata.getDataSize());

            brief.addDataSet(newdata);
          }
        }

        if (brief != null) {
          brief.setID(did);
          brief.updateInfo();
          engine.receivedBriefInfo(brief); 
          register(brief, vid, null);
          brief.setViewerID(vid);
        }

	if (discard)
	  deleteDataSet(dataSet);

      } else if (filter.getStyle() == Filter.MULTIPLE) {

        DataSet[] dss1 = engine.getGUI().iconDisplay.getDataSetBrief(id1).getDataSet();
        DataSet[] dss2 = engine.getGUI().iconDisplay.getDataSetBrief(id2).getDataSet();

        if (dss1.length!=1 || dss2.length!=1) {
          System.out.println("Multi-data merging does not support yet.");
          return;
        }
     
        int noOfInputs = filter.getNoOfArgs();
        filter.clear();

        if (noOfInputs >= 1) {
          System.out.println("Merging with arguments does not support yet.");
        } else {        // if it doesn't need an argument
          filter.assignDataSet(dss1[0]);
          filter.assignDataSet(dss2[0]);
        }

        DataSet newdata;
        try { 
          newdata = (DataSet)filter.performFilter();
        } catch (Exception ee) {
          System.out.println("Client filter action error(Code).");
          return;
        }

        if (newdata == null) {
          System.out.println("Client filter action error(null result).");
          return;
        }

        if (overlay) {
          DataSetBrief dsb = engine.getGUI().iconDisplay.getDataSetBrief(id1);
          dsb.addDataSet(newdata);

          if (dsb.getViewerID() > 0) {
              long vid2 = dsb.getViewerID();
              for (int i=0; i<frameVector.size(); i++) {
                GFrame gframe = (GFrame)frameVector.elementAt(i);
                if (gframe.getID() == vid2) {
                  gframe.notifyRepaint();
                  break;
                }
              }
            }

        } else {
          if (brief == null)
            brief = new DataSetBrief(newdata.getDataName(), 
                                   newdata.getNoOfTimes(), 
                                   filter.getOutDimension(),
                                   newdata.getDataSize());

          brief.addDataSet(newdata);
        }

        if (brief != null) {
          brief.setID(did);
          brief.updateInfo();
          engine.receivedBriefInfo(brief); 
          register(brief, vid, null);
          brief.setViewerID(vid);
        }

        if (discard) 
          System.out.println("Discarding original data for merging does not support yet.");

      }

    } catch (Exception e) {
      System.out.println("Client filter action error(Conversion).");
    }
  }

  /** 
   * register data set to this system
   * @param data data set to visualize
   * @param id expected visulaization window id
   * @param c configurations
   */
  public long register(DataSetBrief brief, long id, Configs c) {
     long fid = -1;
     int dim = brief.getRank();

     switch(dim) {
        case TimeData.ONED:
           fid = register_1d(brief, id, c);
           break;
        case TimeData.TWOD:
           fid = register_2d(brief, id, c);
           break;
        case TimeData.THREED:
           fid = register_3d(brief, id, c);
           break;
        default:
           System.out.println("Unknown type of data set");
     }
     return fid;
  }

  /**
   * Do some work according to the parameter. It will parse the 
   * datum converted to byte array. 
   * @param datum info what to do converted to byte array
   */
  public void doDataWork(byte[] datum) {
    StringTokenizer st = new StringTokenizer(new String(datum), ";");

    // The first token is a command
    int comm = Integer.parseInt(st.nextToken());
    DataSetBrief brief = engine.getGUI().iconDisplay.getDataSetBrief(
                                    Long.parseLong(st.nextToken()));

    switch (comm) {
      case SV2Action.DELETE:
      case SV2Action.CLOSE:
        deleteDataSet(brief);
        break;

      case SV2Action.CUT:
        setClipboard(brief);
        deleteDataSet(brief);
        engine.getGUI().updateMIs(3, true);
        break;
      
      case SV2Action.COPY:
        setClipboard(brief);
        engine.getGUI().updateMIs(3, true);
        break;

      case SV2Action.PASTE:
        pasteAction(brief);
        break;

      case SV2Action.ANIMATE:
        animateAction(brief.getViewerID());
        break;

      case SV2Action.STOP:
        stopAction(brief.getViewerID(), Integer.parseInt(st.nextToken()));
        break;

      case SV2Action.NEXT:
        nextAction(brief.getViewerID());
        break;

      case SV2Action.PREVIOUS:
        previousAction(brief.getViewerID());
        break;

      default:
        GFrame gframe = isExist(brief.getViewerID());
        if (gframe != null) 
          gframe.doActionEventWork(datum);
        else
          System.err.println("Event action error.");
        return;
    }     
  }

  /**
   * delete data set
   */
  public void deleteDataSet(DataSetBrief brief) {
    if (brief == null) return;
    long vid = brief.getViewerID();
    if (vid > 0) {
      for (int i=0; i<frameVector.size(); i++) {
        GFrame gframe = (GFrame)frameVector.elementAt(i);
        if (gframe.getID() == vid) {
          frameVector.removeElement(gframe);
          gframe.closeFrame();
          gframe = null;
          break;
        }
      }
    }

    engine.getGUI().iconDisplay.deleteDataIcon(brief);  
    brief.setClear();
    brief = null;
  }
  
  /**
   * paste action
   */
  public void pasteAction(DataSetBrief brief) {
    if (clipboard.getViewerID() < 0 && brief.getViewerID() > 0) 
      return;

    DataSet[] ds = clipboard.getDataSet();
    if (ds != null) 
      for (int i=0; i<ds.length; i++)
        brief.addDataSet(ds[i]);
    brief.updateInfo();

    if (brief.getViewerID() > 0) {
      long vid = brief.getViewerID();
      for (int i=0; i<frameVector.size(); i++) {
        GFrame gframe = (GFrame)frameVector.elementAt(i);
        if (gframe.getID() == vid) {
          gframe.notifyRepaint();
          break;
        }
      }
    }

    clipboard.setClear();
    clipboard = null;
  }

  /**
   * animate action
   */
  public void animateAction(long vid) {
    if (vid > 0) {
      for (int i=0; i<frameVector.size(); i++) {
        GFrame gframe = (GFrame)frameVector.elementAt(i);
        if (gframe.getID() == vid) {
          gframe.doAnimate();
          return;
        }
      }
    }
  }

  /**
   * next action
   */
  public void nextAction(long vid) {
    if (vid > 0) {
      for (int i=0; i<frameVector.size(); i++) {
        GFrame gframe = (GFrame)frameVector.elementAt(i);
        if (gframe.getID() == vid) {
          gframe.doNext(true);
          return;
        }
      }
    }
  }
  /**
   * previous action
   */
  public void previousAction(long vid) {
    if (vid > 0) {
      for (int i=0; i<frameVector.size(); i++) {
        GFrame gframe = (GFrame)frameVector.elementAt(i);
        if (gframe.getID() == vid) {
          gframe.doPrevious();
          return;
        }
      }
    }
  }

  /**
   * stop action
   */
  public void stopAction(long vid, int time) {
    if (vid > 0) {
      for (int i=0; i<frameVector.size(); i++) {
        GFrame gframe = (GFrame)frameVector.elementAt(i);
        if (gframe.getID() == vid) {
          gframe.doStop(time);
          return;
        }
      }
    }
  }

  // register 1d viewer to this system
  private long register_1d(DataSetBrief brief, long id, Configs c) {
    if (viewer_opt == SV2Engine.VIEWER_BUILTIN) {

      if (c == null) 
        c = config.getConfig1D();

      DataSet[] data = brief.getDataSet();
      TimeData1D td = (TimeData1D)data[0].getTimeData(0);
      int dataType = td.getDataType();

      GFrame1D viewer = null;
      if (dataType == TimeData1D.POINT)
        viewer = new GFPoint(brief, c);
      else if (dataType == TimeData1D.CONNECTIVITY)
        viewer = new GFCon(brief, c);

      frameVector.addElement(viewer);
      if (id > 0) viewer.setID(id);
      else viewer.setID(System.currentTimeMillis());

      viewer.setSize(400, 400);
      viewer.show();    
      return viewer.getID();
    } else {
      System.out.println("Visualization other than Java3D is not supported yet.");
    } 

    return -1; 
  }

  // register 2d viewer to this system
  private long register_2d(DataSetBrief brief, long id, Configs c) {
    if (viewer_opt == SV2Engine.VIEWER_BUILTIN) {

      if (c == null) 
        c = config.getConfig2D();

      GFrame2D viewer = new GFrame2D(brief, c);
      frameVector.addElement(viewer);
      if (id > 0) viewer.setID(id);
      else viewer.setID(System.currentTimeMillis());

      viewer.setSize(400, 400);
      viewer.show();    
      return viewer.getID();

    } else {
      System.out.println("Visualization other than Java3D is not supported yet.");
    } 
    return -1; 
  }

  // register 3d viewer to this system
  private long register_3d(DataSetBrief brief, long id, Configs c) {
    if (viewer_opt == SV2Engine.VIEWER_BUILTIN) {

      if (c == null) 
        c = config.getConfig3D();

      GFrame3D viewer = new GFrame3D(brief, c);
      frameVector.addElement(viewer);
      if (id > 0) viewer.setID(id);
      else viewer.setID(System.currentTimeMillis());

      viewer.setSize(400, 400);
      viewer.show();    
      return viewer.getID();


    } else {
      System.out.println("Visualization other than Java3D is not supported yet.");
    } 
    return -1; 
  }

  /** 
   * register volume data set window to this system
   * @param data data set to visualize
   * @param file image file directory path
   * @param arr image byte array
   * @param vars display options
   * @param id expected visulaization window id
   */
  public long register(DataSetBrief brief, File path , byte[] arr, 
                                              double[] vars, long id) {

    if (viewer_opt == SV2Engine.VIEWER_BUILTIN) {

      GFrameVol viewer = 
               new GFrameVol(brief, path, arr, vars, config.getConfig3D());
      frameVector.addElement(viewer);
      if (id > 0) viewer.setID(id);
      else viewer.setID(System.currentTimeMillis());

      viewer.pack();
      viewer.show();    
      return viewer.getID();

    } else {
      System.out.println("Visualization other than Java3D is not supported yet.");
    } 
    return -1; 
  }

  /**
   * return viewer option
   */
  public int getViewerOption() {
    return viewer_opt;
  }

  /**
   * return default configuration
   */
  public ClientConfig getConfig() {
    return config;
  }


  class MergeHandle {
    DataSetBrief target;
    Filter filter;
    boolean overlay = false;
    boolean discard = false;

    MergeHandle(Filter filter, DataSetBrief target, 
             boolean overlay, boolean discard) {
      
      this.target = target;
      this.filter = filter;
      this.overlay = overlay;
      this.discard = discard;
    }
  } // end of class MergeHandle

} // end of class

