package sv.viewer.builtin;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.geom.AffineTransform;
import java.util.Vector;
import sv.viewer.*;
import sv.data.*;
import sv.config.*;
import sv.gui.*;

public abstract class GCanvas1D extends Canvas 
                              implements MouseListener, MouseMotionListener {

  // fried class of this 
  protected GFrame1D gframe;

  // configuration of the friend class
  protected Config1D config;

  // data set
  protected DataSetBrief brief;

  // display area size
  protected Dimension canvasSize;

  // mouse related work
  public static final int CLOSEST_POINT_MODE             = 1;
  public static final int DATA_AREA_RESIZE_MODE          = 2;
  public static final int LABEL_MOVE_MODE                = 3;
  public static final int LEGEND_MOVE_MODE               = 4;
  public static final int PANNING_MODE                   = 5;
  public static final int ZOOMING_MODE                   = 6;
  public static final int DRAWING_MODE                   = 7;

  // mouse mode 
  protected int mouseMode = CLOSEST_POINT_MODE;

  // canvas size of point data area 
  protected Rectangle dataArea;

  // mouse click point
  protected int clickX = -1, clickY = -1;

  // min and max value of given time slot
  protected float xmin, xmax, ymin, ymax;

  // closest point
  protected boolean drawClosest = false;
  protected float[] closest;
  protected String displayStr;

  // buffered images
  protected BufferedImage offImg;
  protected BufferedImage pdaImg;
  protected Graphics2D g2;
  protected Graphics2D pdaG2;

  // if user entered min and max, it will turn on
  protected boolean userMinMax = false;

  // pdaImg size for fast calculation
  protected Dimension imgSize;

  // printing
  protected boolean printing = false;        // printing flag
  protected int printwidth, printheight;     // print size

  // font size and line width
  protected int fontsize = 12;
  protected float linewidth = 1.0f;
 
  // about the label, legend, data area corners
  boolean isXLabelRotate = false; 
  boolean isYLabelRotate = false; 
  Point xLabelPoint = new Point();
  Point yLabelPoint = new Point();
  int selectedLabel;
  Rectangle legendRect = new Rectangle();
  Rectangle legendCenter;
  int selectedCorner;
  Rectangle pdaCorners[];

  // Axis
  Axis axis;

  public GCanvas1D(GFrame1D gframe, Config1D cfg) {
    this.gframe = gframe;
    config = cfg;

    canvasSize = new Dimension(-1, -1);
    pdaCorners = new Rectangle[2];
    axis = new Axis();

    createKeyAdapter(); 
    addMouseListener(this);
    addMouseMotionListener(this);
  }

  /**
   * overrided paint method
   */
  public void paint(Graphics g) {
    Dimension d = getSize();
    brief = gframe.getDataSet();

    if (!userMinMax) {
      float[] minmax = gframe.getMinMax();
      xmin = minmax[0];
      xmax = minmax[1];
      ymin = minmax[2];
      ymax = minmax[3];
    }

    if (offImg == null || 
          d.width != canvasSize.width || d.height != canvasSize.height) {

      offImg = (BufferedImage)createImage(d.width, d.height);
      g2 = offImg.createGraphics();
      if (!changeArea()) return;
    }

    printing = g instanceof PrintGraphics;

    if (printing) {
      // modifies variables
      Graphics2D savedgc = g2;
      g2 = (Graphics2D)g;
      
      Dimension pagedimension = 
                 ((PrintGraphics)g).getPrintJob().getPageDimension();
      
      printwidth = pagedimension.width;
      printheight = canvasSize.height * canvasSize.width / canvasSize.width;

      if (printheight > pagedimension.height) {
        printheight = pagedimension.height;
        printwidth = canvasSize.width * canvasSize.height / canvasSize.height;
      }

      // erase whole canvas
      g2.clearRect(0, 0, printwidth, printheight);
      g2.setColor(config.getBGColor());
      g2.fillRect(0, 0, printwidth, printheight);

      ////////////////////////////////////////////
      // SHOULD DO MORE WORKS HERE
      ////////////////////////////////////////////

      g2 = savedgc;
      return;
    }
  
    // erase the whole canvas
    g2.clearRect(0, 0, d.width, d.height);
    g2.setColor(config.getBGColor());
    g2.fillRect(0, 0, d.width, d.height);

    // if the canvas size is too small
    if (d.width < 50 || d.height < 50) {
      g2.setColor(Color.red);
      g2.drawString("Too small canvas.", 5, (int)(d.height/2));
      return;
    }

    // prepare the graphics
    int x = config.getPanX();
    int y = config.getPanY();
    pdaG2.setClip(x, y, dataArea.width, dataArea.height);

    // draw point data
    drawData();

    g2.drawImage(pdaImg, dataArea.x, dataArea.y, 
             dataArea.x+dataArea.width, dataArea.y+dataArea.height,
             x, y, x+dataArea.width, y+dataArea.height, this);

    // draw labels and title
    drawMisc();

    if (offImg != null)
      g.drawImage(offImg, 0, 0, this);
  }

  /**
   * overrided update method
   */
  public void update(Graphics g) {
    paint(g);
  }

  /**
   * change the canvas area to be drawn the data set
   */
  public boolean changeArea() {
    canvasSize = getSize();
    Dimension d = canvasSize;

    if (offImg == null) return false;

    g2.setFont(new Font("TimesRoman", Font.PLAIN, fontsize));
    FontMetrics fm = g2.getFontMetrics();
    int tickhgt = fm.getHeight();
    int lblhgt = 20;

    // subtract the parts for labels and title to calculate dataArea size
    if (config.isTickMarkDraw()) {

      if (config.isLabelDraw()) {          // ticker and label
        dataArea = new Rectangle(60+lblhgt, 20+tickhgt, 
                            d.width-lblhgt-80, d.height-40-2*tickhgt-lblhgt);

        xLabelPoint.x = (dataArea.x + dataArea.width) / 2;
        xLabelPoint.y = d.height - lblhgt - 10;
        yLabelPoint.x = 10;
        yLabelPoint.y = (dataArea.y + dataArea.height) / 2;

      } else {                            // ticker only               
        dataArea = new Rectangle(60, 20+tickhgt, 
                            d.width-80, d.height-40-2*tickhgt);
      }
      axis.setTickMode(true);

    } else {

      if (config.isLabelDraw()) {         // label only
        dataArea = new Rectangle(20+lblhgt, 20+tickhgt, 
                            d.width-lblhgt-40, d.height-40-2*tickhgt-lblhgt);

        xLabelPoint.x = (dataArea.x + dataArea.width) / 2;
        xLabelPoint.y = d.height - lblhgt - 10;
        yLabelPoint.x = 10;
        yLabelPoint.y = (dataArea.y + dataArea.height) / 2;

      } else {                            // no label and ticker
        dataArea = new Rectangle(20, 20+tickhgt,
                            d.width-40, d.height-40-2*tickhgt);
      }
      axis.setTickMode(false);
    }

    if (config.isLegendDraw()) {
      dataArea = new Rectangle(dataArea.x, dataArea.y+4*tickhgt,
                  dataArea.width, dataArea.height-4*tickhgt);
      legendCenter = null;
    }

    createPDAGraphics();
    return true;
  }


  /**
   * set mouse mode
   */
  public void setMouseMode(int mode) {
    mouseMode = mode;

    if (mouseMode == PANNING_MODE)
      setCursor(new Cursor(Cursor.MOVE_CURSOR));
    else if (mouseMode == ZOOMING_MODE)
      setCursor(new Cursor(Cursor.HAND_CURSOR));
    else
      setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
  }

  /**
   * change font size
   * @param b if true , increase the font size
   */
  public void changeFont(boolean b) {
    if (b)
      fontsize += 2;
    else {
      if (fontsize < 2) return;
      else fontsize -= 2;
    }
    changeArea();
    repaint();
  }

  /**
   * change line thickness
   * @param b if true , increase the line width
   */
  public void changeLineThick(boolean b) {
    if (b)
      linewidth = linewidth + 0.4f;
    else {
      if (linewidth <= 1.0) return;
      else linewidth = linewidth - 0.4f;
    }
    repaint();
  }


  //////////////////////////////////////////////////////////
  // mouse and key events
  //////////////////////////////////////////////////////////

  // create the key adapter
  protected void createKeyAdapter() {
    KeyListener l = new KeyAdapter() {
      public void keyTyped(KeyEvent e) {
      }
    };
    this.addKeyListener(l);
  }

  /**
   * Invoked when the mouse has been clicked on a component.
   */
  public void mouseClicked(MouseEvent e) {
    switch (mouseMode) {
      case CLOSEST_POINT_MODE:
        return;

      case ZOOMING_MODE:
        float factor = config.getZoomFactor();
        if (e.isControlDown()) {   // unzoom mode
          if (factor < 1.5f) {
            gframe.message("Already reached minimum unzoom level", true);
            return;
          } 
          factor = factor / 1.5f;
        } else 
          factor = factor * 1.5f;

        config.setZoomFactor(factor);
        createPDAGraphics();

        if (config.getPanX()+dataArea.width > imgSize.width) 
          config.setPanX(imgSize.width - dataArea.width);

        if (config.getPanY()+dataArea.height > imgSize.height) 
           config.setPanY(imgSize.height - dataArea.height);

        gframe.message("Zoom Factor: " + config.getZoomFactor(), false);
        break;
    }

    repaint();
  }

  /**
   * Invoked when a mouse button has been pressed on a component.
   */
  public void mousePressed(MouseEvent e) {
    clickX = e.getX();
    clickY = e.getY();

    int x = clickX;
    int y = clickY;

    if (mouseMode == DATA_AREA_RESIZE_MODE) {
      for (int i=0; i<2; i++) {
        if (pdaCorners[i].contains(x, y)) {
          selectedCorner = i;
          return;
        }
      }
      mouseMode = CLOSEST_POINT_MODE;
    }

    if (mouseMode == LEGEND_MOVE_MODE) {
      if (legendCenter.contains(x, y))
        return;
      mouseMode = CLOSEST_POINT_MODE;
    }

    if (mouseMode == LABEL_MOVE_MODE) {
      Rectangle xrect = new Rectangle(xLabelPoint.x, 
                                      xLabelPoint.y, 20, 20);
      Rectangle yrect = new Rectangle(yLabelPoint.x, 
                                      yLabelPoint.y, 20, 20);

      if (xrect.contains(x, y)) {
        selectedLabel = 1;
        return;
      }
 
      if (yrect.contains(x, y)) {
        selectedLabel = 2;
        return;
      }
      mouseMode = CLOSEST_POINT_MODE;
    }
  }

  /**
   * Invoked when a mouse button has been released on a component.
   */
  public void mouseReleased(MouseEvent e) {
    clickX = -1;
    clickY = -1;

    int x = e.getX() - dataArea.x;
    int y = e.getY() - dataArea.y;

    Rectangle rect = new Rectangle(dataArea.width+5, dataArea.height+5);

    if (rect.contains(x, y)) {
      float scalex = (float)dataArea.width / (xmax-xmin);
      float scaley = (float)dataArea.height / (ymax-ymin);

      float clickedx = (float)x/scalex + xmin;
      float clickedy = (float)(dataArea.height-y)/scaley + ymin;

      closest = getClosestPoint(clickedx, clickedy);
      drawClosest = true;
      displayStr =  "(" + String.valueOf(clickedx) + ", " + 
                          String.valueOf(clickedx);
      repaint();
    }
  }

  /**
   * Invoked when the mouse enters a component.
   */
  public void mouseEntered(MouseEvent e) {
  }

  /**
   * Invoked when the mouse exits a component.
   */
  public void mouseExited(MouseEvent e) {
  }

  /** 
   * Invoked when a mouse button is pressed on a component and then dragged.
   * Mouse drag events will continue to be delivered to the component 
   * where the first originated until the mouse button is released 
   * (regardless of whether the mouse position is within the bounds of the 
   * component).
   */
  public void mouseDragged(MouseEvent e) {
    int x = e.getX();
    int y = e.getY();

    Rectangle rect;

    if (mouseMode == DATA_AREA_RESIZE_MODE) {
      switch (selectedCorner) {
        case 0:
          rect = new Rectangle(x, y, 
                 dataArea.x+dataArea.width-x, dataArea.y+dataArea.height- y);
          if (rect.width > 0 && rect.height > 0)
              dataArea = rect;
          break;

        case 1:
          int hgt = dataArea.y - y;
          int wid = dataArea.x - x;

          if (hgt < 0)
              hgt *= -1;
          if (wid < 0)
              wid *= -1;

          rect = new Rectangle(dataArea.x, dataArea.y, wid, hgt);

          if (rect.width > 0 && rect.height > 0) 
              dataArea = rect;

          break;

        default:
            return;
      }
      createPDAGraphics();

    } else if (mouseMode == LABEL_MOVE_MODE) {
      if (selectedLabel == 1) {
        xLabelPoint.x = x;
        xLabelPoint.y = y;
      } else if (selectedLabel == 2) {
        yLabelPoint.x = x;
        yLabelPoint.y = y;
      }

    } else if (mouseMode == LEGEND_MOVE_MODE) {
      Rectangle r1 = legendRect;
      Rectangle r2 = new Rectangle(x-r1.width/2, y-r1.height/2, 
                                                   r1.width, r1.height);
      int midx = Math.round((float)r2.x + (float)r2.width / 2.0f);
      int midy = Math.round((float)r2.y + (float)r2.height / 2.0f);
      legendCenter = new Rectangle(midx-10, midy-10, 20, 20);

    } else if (mouseMode == PANNING_MODE) {
      if (!panCheck(x, y)) {
        gframe.message("Can't pan further.", true);
        return;
      }
    }

    repaint();
  }

  /**
   * Invoked when the mouse button has been moved on a component (with no
   *  buttons no down).
   */
  public void mouseMoved(MouseEvent e) {

  }

  // panning work
  protected boolean panCheck(int x, int y) {
    int diffx = clickX - x;
    int diffy = clickY - y;

    int newPanX = config.getPanX() + diffx;
    int newPanY = config.getPanY() + diffy;
    
    if (diffx != 0) {
      if (imgSize.width <= dataArea.width || newPanX < 0 || 
                             newPanX+dataArea.width >= imgSize.width)
        return false; 
    } 

    if (diffy != 0) {
      if (imgSize.height <= dataArea.height || newPanY < 0 || 
                             newPanY+dataArea.height >= imgSize.height)
        return false;
    }
   
    clickX = x;
    clickY = y;

    config.setPanX(newPanX);
    config.setPanY(newPanY);
    return true;  
  }

  // create the pda graphics
  protected void createPDAGraphics() {
    imgSize = new Dimension(
                (int)((float)dataArea.width*config.getZoomFactor()),
                (int)((float)dataArea.height*config.getZoomFactor()));

    pdaCorners[0] = new Rectangle(dataArea.x-10, dataArea.y-10, 20, 20);
    pdaCorners[1] = new Rectangle(dataArea.x+dataArea.width-10, 
                                  dataArea.y+dataArea.height-10, 20, 20);

    pdaImg = (BufferedImage)createImage(imgSize.width, imgSize.height);

    if (pdaImg != null) 
      pdaG2 = pdaImg.createGraphics();

    pdaG2.setRenderingHint(RenderingHints.KEY_RENDERING,
                               RenderingHints.VALUE_RENDER_QUALITY);
  }

   
  // draw point data
  protected abstract void drawData();

 // draw labels and title
  protected void drawMisc() {
    g2.setColor(config.getFGColor());
    int panX = config.getPanX();
    int panY = config.getPanY();

    //////////////////////////////////////////////////////////////
    // display labels
    if (config.isLabelDraw()) {
      int hgt = g2.getFontMetrics().getHeight();

      // display x-axis label
      if (isXLabelRotate) {
        AffineTransform at = new AffineTransform();
        at.translate(xLabelPoint.x, xLabelPoint.y);
        at.rotate(-Math.PI / 2.0);	
        g2.transform(at);
        g2.drawString(config.getXLabel(), 0, 0);
        g2.setTransform(new AffineTransform());

      } else 
         g2.drawString(config.getXLabel(), xLabelPoint.x, xLabelPoint.y);

      // display y-axis label
      if (isYLabelRotate) {
        AffineTransform at = new AffineTransform();
        at.translate(yLabelPoint.x, yLabelPoint.y);
        at.rotate(-Math.PI / 2.0);	
        g2.transform(at);
        g2.drawString(config.getYLabel(), 0, 0);
        g2.setTransform(new AffineTransform());

      } else 
         g2.drawString(config.getYLabel(), yLabelPoint.x, yLabelPoint.y);

      if (mouseMode == LABEL_MOVE_MODE) {
        g2.setColor(Color.lightGray);
        g2.fillRect(xLabelPoint.x, xLabelPoint.y, 20, 20);
        g2.fillRect(yLabelPoint.x, yLabelPoint.y, 20, 20);
        g2.setColor(config.getFGColor());
      }
    }

    //////////////////////////////////////////////////////////////
    // display legend
    if (config.isLegendDraw()) {
      calcLegendArea();

      int x = legendRect.x;
      int y = legendRect.y;
      int w = legendRect.width;
      int h = legendRect.height;

      // draw boundary rectangle
      g2.setColor(config.getBGColor());
      g2.fillRect(x, y, w, h);
      g2.setColor(config.getFGColor());
      g2.drawRect(x, y, w, h);

      DataSet[] data = brief.getDataSet();
      int gap = h / data.length;
      int mpos = x + 20;                            // marker position
      FontMetrics fm = g2.getFontMetrics();

      int f = (int)(2.0f * linewidth);
      int f2 = 2 * f;
      for (int i=0; i<data.length; i++) {
        int pos = y + i*gap + gap/2;                // line position
        int lblp = y + i*gap + 5 + fm.getMaxAscent();   // label position

        g2.setColor(data[i].getLineColor());
        g2.drawLine(x+5, pos, x+35, pos);

        switch (data[i].getMarkerType()) {
          case MarkerChooser.CIRCLEMARKER:
            g2.drawOval(mpos-f, pos-f, f2, f2);
            break;

          case MarkerChooser.SQUAREMARKER:
            g2.drawRect(mpos-f, pos-f, f2, f2);
            break;

          case MarkerChooser.DIAMONDMARKER:
            g2.drawLine(mpos, pos-f, mpos-f, pos);
            g2.drawLine(mpos-f, pos, mpos, pos+f);
            g2.drawLine(mpos, pos+f, mpos+f, pos);
            g2.drawLine(mpos+f, pos, mpos, pos-f);
            break;

          case MarkerChooser.CIRCLEFILLMARKER:
            g2.fillOval(mpos-f, pos-f, f2, f2);
            break;

          case MarkerChooser.SQUAREFILLMARKER:
            g2.fillRect(mpos-f, pos-f, f2, f2);
            break;

          case MarkerChooser.DIAMONDFILLMARKER:
            int[] xpts = {mpos,  mpos-f, mpos, mpos+f };
            int[] ypts = {pos-f, pos, pos+f, pos };

            g2.fillPolygon(xpts, ypts, 4);
            break;

          case MarkerChooser.CROSSMARKER:
            g2.drawLine(mpos, pos-f, mpos, pos+f);
            g2.drawLine(mpos-f, pos, mpos+f, pos);
            break;

          case MarkerChooser.XMARKER:
            g2.drawLine(mpos-f, pos-f, mpos+f, pos+f);
            g2.drawLine(mpos+f, pos-f, mpos-f, pos+f);
            break;
        }

        g2.setColor(config.getFGColor());
        g2.drawString(data[i].getDataName(), x + 40, lblp);
      }

      if (mouseMode == LEGEND_MOVE_MODE) {
        g2.setColor(Color.lightGray);
        g2.fillRect(legendCenter.x, legendCenter.y, 
                              legendCenter.width, legendCenter.height);
        g2.setColor(config.getFGColor());
      }
    }

    //////////////////////////////////////////////////////////////
    // draw Axis
    if (imgSize.width > 0 && imgSize.height > 0) {
      float xgap = (xmax - xmin) / (float)imgSize.width;
      float ygap = (ymax - ymin) / (float)imgSize.height;

      axis.setAxisBounds(dataArea);
      axis.setAxisMinMax(
              xgap * (float)panX + xmin,
              xgap * (float)(panX+dataArea.width) + xmin,
              ygap * (float)(imgSize.height-panY-dataArea.height) + ymin,
              ygap * (float)(imgSize.height-panY) + ymin);
      axis.paint(g2);
    }

    //////////////////////////////////////////////////////////////
    // display time
    if (config.isDisplayTimeMode()) {
      FontMetrics fm = g2.getFontMetrics(g2.getFont());
      g2.drawString("Time: " + gframe.getCurrent(), 5, fm.getMaxAscent()+10);
    }

    if (mouseMode == DATA_AREA_RESIZE_MODE) {
      g2.setColor(Color.lightGray);
      g2.fillRect(pdaCorners[0].x, pdaCorners[0].y, 
                         pdaCorners[0].width, pdaCorners[0].height);
      g2.fillRect(pdaCorners[1].x, pdaCorners[1].y, 
                         pdaCorners[1].width, pdaCorners[1].height);
    }
  }

  // Find the closest data point to the cursor
  protected float[] getClosestPoint(float x, float y) {
    float a[] = new float[3];
    float distsq = -1.0f;
    float data[] = {0.0f, 0.0f};

    DataSet[] ds = brief.getDataSet();
    for (int i=0; i<ds.length; i++) {
      int sz = ds[i].getNoOfTimes();
      for (int j=0; j<sz; j++) {
        if (ds[i].getTimeValue(j).getAbsTime() == gframe.getCurrent()) {
          TimeData1D td = (TimeData1D)ds[i].getTimeData(j);
          a = td.getClosestPoint(x,y);
          if (distsq < 0.0f || distsq > a[2]) {
             data[0] = a[0];
             data[1] = a[1];
             distsq  = a[2];
          }
        }
      }
    }
    return data;
  }

  // calculate the legend area
  protected void calcLegendArea() {
    DataSet[] data = brief.getDataSet();
    int size = data.length;

    FontMetrics fm = g2.getFontMetrics();
    Dimension d = getSize();

    int hgt = (fm.getHeight()+10) * size;
    int wid = 0;

    for (int i=0; i<size; i++) {
      int w = fm.stringWidth(data[i].getDataName());
      if (wid < w)
        wid = w;
    }
    wid += 50;

    if (legendCenter == null) {
      legendRect = new Rectangle(d.width-wid-10, 10, wid, hgt);

      int midx = 
          Math.round((float)legendRect.x + (float)legendRect.width/2.0f);
      int midy = 
          Math.round((float)legendRect.y + (float)legendRect.height / 2.0f);
      legendCenter = new Rectangle(midx-10, midy-10, 20, 20);
      return;
    }

    legendRect = 
          new Rectangle(legendCenter.x-wid/2, legendCenter.y-hgt/2, wid, hgt);
  }

} // end of class


