package sv.viewer.builtin;

import java.awt.*;
import java.util.*;
import java.io.*;
import java.awt.event.*;
import com.sun.java.swing.*;
import com.sun.java.swing.border.*;
import sv.engine.*;
import sv.viewer.*;
import sv.config.*;
import sv.data.*;
import sv.gui.*;
import sv.util.*;

/** 
 * Define a frame for 1d data
 */
public abstract class GFrame1D extends GFrame {

  // canvas to draw the graph
  protected GCanvas1D canvas;

  // global min and max
  protected float globalxmin = Float.MAX_VALUE, globalxmax = Float.MIN_VALUE, 
                globalymin = Float.MAX_VALUE, globalymax = Float.MIN_VALUE;

  // menu items
  protected MenuItem splitMI, byDataSetMI, 
                    changeLabelMI, moveLabelMI, changeLegendMI, moveLegendMI, 
                    changeTickMI;

  protected CheckboxMenuItem displayLabelMI, displayLegendMI, drawTickMI, 
                           connectMI, displayTimeMI, globalMinMaxMI;

  /**
   * constructor
   * @param brief initial data set
   * @param cfg configuration of this window
   */
  public GFrame1D(DataSetBrief brief, Configs cfg) {
    super(brief, cfg);
    calcGlobalMinMax();  
  
    WindowListener l = new WindowAdapter() {
      public void windowClosing(WindowEvent e) {
      }
    };
    this.addWindowListener(l);
  } 

  /**
   * display message on the staus bar
   */
  public void message(String msg, boolean critical) {
    status.display(msg, critical);
  }

  /**
   * return the type of data set in this frame
   */
  public int getDataType() {
    DataSet[] data = dataSet.getDataSet();
    TimeData1D td = (TimeData1D)data[0].getTimeData(0);
    return td.getDataType();
  }

  /**
   * get min and max
   */
  public float[] getMinMax() {
    float[] minmax = new float[4];
    minmax[0] = Float.MAX_VALUE;
    minmax[1] = Float.MIN_VALUE;
    minmax[2] = Float.MAX_VALUE;
    minmax[3] = Float.MIN_VALUE;

    if (globalMinMaxMI.getState()) {
      minmax[0] = globalxmin;
      minmax[1] = globalxmax;
      minmax[2] = globalymin;
      minmax[3] = globalymax;
    } else {
      DataSet[] data = dataSet.getDataSet();
      float time = timeValues[currentTime];
      for (int i=0; i<data.length; i++) {
        int sz = data[i].getNoOfTimes();
        for (int j=0; j<sz; j++) 
          if (data[i].getTimeValue(j).getAbsTime() == time) {
            TimeData1D td = (TimeData1D)data[i].getTimeData(j);
            if (td.getxmin() < minmax[0]) minmax[0] = td.getxmin();
            if (td.getxmax() > minmax[1]) minmax[1] = td.getxmax();
            if (td.getymin() < minmax[2]) minmax[2] = td.getymin();
            if (td.getymax() > minmax[3]) minmax[3] = td.getymax();
            break;
	  }
      }
    }
    return minmax;
  }

  /**
   * change color map info
   */
  public void changeColorMap(String[] args) {
    String comm = String.valueOf(SV2Action.COLORMAP_CHANGE) + ";" +
                      String.valueOf(getDataSet().getID());
    for (int i=0; i<args.length; i++)
      comm = comm + ";" + args[i];
    catchActionEvent(comm);
  }

  /**
   * repaint
   */
  public void notifyRepaint() {
    calcGlobalMinMax();
    canvas.repaint();
  }

  /**
   * do action event
   */
  public void doActionEventWork(byte[] datum) {
    StringTokenizer st = new StringTokenizer(new String(datum), ";");

    // The first token is a command
    int comm = Integer.parseInt(st.nextToken());
    st.nextToken();   // trash it

    DataSet[] data;
    boolean flag;
    int sz, r, g, b;
    
    switch (comm) {
      case SV2Action.CONNECT_POINT:
        flag = Boolean.valueOf(st.nextToken()).booleanValue();
        ((Config1D)config).setConnectPoints(flag);
        connectMI.setState(flag);
        canvas.repaint();
        break;

      case SV2Action.TICKMARKS:
        flag = Boolean.valueOf(st.nextToken()).booleanValue();
        config.setTickMarkDraw(flag);
        drawTickMI.setState(flag);
        canvas.changeArea(); 
        canvas.repaint();
        break;

      case SV2Action.DISPLAY_LABEL:
        flag = Boolean.valueOf(st.nextToken()).booleanValue();
        config.setLabelDraw(flag);
        displayLabelMI.setState(flag);
        canvas.changeArea(); 
        canvas.repaint();
        break;

      case SV2Action.DISPLAY_LEGEND:
        flag = Boolean.valueOf(st.nextToken()).booleanValue();
        config.setLegendDraw(flag);
        displayLegendMI.setState(flag);
        canvas.changeArea(); 
        canvas.repaint();
        break;

      case SV2Action.DISPLAY_TIME:
        flag = Boolean.valueOf(st.nextToken()).booleanValue();
        ((Config1D)config).setDisplayTimeMode(flag);
        displayTimeMI.setState(flag);
        canvas.repaint();
        break;

      case SV2Action.BACKGROUND_COLOR:
        r = Integer.parseInt(st.nextToken());
        g = Integer.parseInt(st.nextToken());
        b = Integer.parseInt(st.nextToken());
        config.setBGColor(new Color(r, g, b));
        canvas.repaint();
        break;

      case SV2Action.FOREGROUND_COLOR:
        r = Integer.parseInt(st.nextToken());
        g = Integer.parseInt(st.nextToken());
        b = Integer.parseInt(st.nextToken());
        config.setFGColor(new Color(r, g, b));
        canvas.repaint();
        break;

      case SV2Action.GLOBALMINMAX:
        flag = Boolean.valueOf(st.nextToken()).booleanValue();
        globalMinMaxMI.setState(flag);
        notifyRepaint();
        break;

      case SV2Action.LEGEND_CHANGE:
        sz = Integer.parseInt(st.nextToken());
        data = dataSet.getDataSet();
        if (sz < data.length * 3)
          return;
        
        for (int i=0; i<data.length; i++) {
          data[i].setDataName(st.nextToken());
          int rgb = Integer.parseInt(st.nextToken());
          data[i].setLineColor(new Color(rgb));
          data[i].setMarkerType(Integer.parseInt(st.nextToken()));
        }
        canvas.repaint();
        break;
   
      case SV2Action.LABEL_CHANGE:
        sz = Integer.parseInt(st.nextToken());
        ((Config1D)config).setXLabel(st.nextToken());
        ((Config1D)config).setYLabel(st.nextToken());
        canvas.isXLabelRotate = Boolean.valueOf(st.nextToken()).booleanValue();
        canvas.isYLabelRotate = Boolean.valueOf(st.nextToken()).booleanValue();
        canvas.repaint();
        break;

    }
  }

  /**
   * chnage legend
   */
  public void legendChange(String[] arg) {
    String comm = String.valueOf(SV2Action.LEGEND_CHANGE) + ";" +
                  String.valueOf(getDataSet().getID());
                  String.valueOf(arg.length);
    for (int i=0; i<arg.length; i++)
      comm = comm + ";" + arg[i];

    catchActionEvent(comm);
  }

  /**
   * update data set
   */
  public abstract void updateDataSet();

  /**
   * animate 
   */
  public void doAnimate() {
    if (animator == null) {
      animator = new Animator(this, animationSpeed);
      animator.start();
    } else 
      System.out.println("Animation thread was corrupted ...");

    animateMI.setEnabled(false);
    nextMI.setEnabled(false);
    prevMI.setEnabled(false);
    stopMI.setEnabled(true);
  }

  /**
   * next
   */
  public void doNext(boolean b) {

    currentTime++;
    if (currentTime == timeValues.length)
      currentTime = 0;

    if (b) 
      canvas.repaint();
  }

  /**
   * previous 
   */
  public void doPrevious() {

    currentTime--;
    if (currentTime == -1)
       currentTime = timeValues.length-1;

    canvas.repaint();
  }

  /**
   * stop 
   */
  public void doStop(int pos) {
    if (animator != null) {
      animator.stop();
      animator = null;
    }
    animateMI.setEnabled(true);
    stopMI.setEnabled(false);
    nextMI.setEnabled(true);
    prevMI.setEnabled(true);
  }


  /**
   * close this visualization window
   */
  public void closeFrame() {

    //////////////////////////////////////////////////////
    // KILL ATTACHED THREADS

    canvas = null;
    this.dispose();    
  }

  /**
   * write misc information into the file
   */
  public void writeAux(DataOutputStream out) throws IOException {
    out.writeByte(0);
  }


  // calc global min and max
  protected void calcGlobalMinMax() {

    DataSet[] data = dataSet.getDataSet();
    for (int i=0; i<data.length; i++) {
      int sz = data[i].getNoOfTimes();
      for (int j=0; j<sz; j++) {
        TimeData1D td = (TimeData1D)data[i].getTimeData(j);
        if (td.getxmin() < globalxmin) globalxmin = td.getxmin();
        if (td.getxmax() > globalxmax) globalxmax = td.getxmax();
        if (td.getymin() < globalymin) globalymin = td.getymin();
        if (td.getymax() > globalymax) globalymax = td.getymax();
      }
    }
  }

} // end of class

